# Copyright (C) 2025 Alexander Vanhee, tfuxu, kjozsa
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from typing import Optional
from gi.repository import Gtk, Gdk, GObject, GLib
from gradia.overlay.drawing_actions import DrawingMode
from gradia.ui.widget.quick_color_picker import QuickColorPicker, SimpleColorPicker
from gradia.ui.widget.drawing_tools_grid import DrawingToolsGrid
from gradia.backend.tool_config import ToolOption, ToolOptionsManager, ToolConfig
from gradia.ui.widget.font_dropdown import FontDropdown
from gradia.constants import rootdir

@Gtk.Template(resource_path=f"{rootdir}/ui/drawing_tools_group.ui")
class DrawingToolsGroup(Gtk.Box):
    __gtype_name__ = "GradiaDrawingToolsGroup"

    size_scale = Gtk.Template.Child()
    color_picker = Gtk.Template.Child()
    extra_stack = Gtk.Template.Child()
    extra_stack_revealer = Gtk.Template.Child()
    fill_0 = Gtk.Template.Child()
    fill_1 = Gtk.Template.Child()
    fill_2 = Gtk.Template.Child()
    outline_1 = Gtk.Template.Child()
    outline_2 = Gtk.Template.Child()
    drawing_tools_grid = Gtk.Template.Child()
    font_dropdown = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.tool_manager = ToolOptionsManager()
        self.current_tool_config: Optional[ToolConfig] = None
        self.current_tool_option: Optional[ToolOption] = None
        self._updating_ui = False
        self._scroll_accum = 0.0

        self.is_temp_editing = False
        self.temp_editing_tool_option: Optional[ToolOption] = None
        self.temp_editing_tool_config: Optional[ToolConfig] = None

        self.connect("realize", self._on_realize)

    def _on_realize(self, *args):
        self._scroll_controller = Gtk.EventControllerScroll.new(
            Gtk.EventControllerScrollFlags.VERTICAL |
            Gtk.EventControllerScrollFlags.HORIZONTAL
        )
        self._scroll_controller.connect("scroll", self._on_scroll)
        self.get_root().add_controller(self._scroll_controller)
        self.get_root().drawing_overlay.connect('selection-changed', self.on_overlay_selected_changed)

    def _get_active_tool_option(self) -> Optional[ToolOption]:
        return self.temp_editing_tool_option if self.is_temp_editing else self.current_tool_option

    def _get_active_tool_config(self) -> Optional[ToolConfig]:
        return self.temp_editing_tool_config if self.is_temp_editing else self.current_tool_config

    def apply_changes(self):
        if self.is_temp_editing:
            self.get_root().drawing_overlay.queue_draw()
        else:
            window = self.get_root()
            if window:
                action = window.lookup_action("tool-option-changed")
                if action and self.current_tool_option:
                    data_json = self.current_tool_option.serialize()
                    param = GLib.Variant('s', data_json)
                    action.activate(param)

    @Gtk.Template.Callback()
    def on_tool_changed(self, grid: DrawingToolsGrid, tool_config: ToolConfig):
        self.get_root().drawing_overlay.selected_action = None
        self.get_root().drawing_overlay.queue_draw()

        self.current_tool_config = tool_config
        self.current_tool_option = self.tool_manager.get_tool(tool_config.mode)
        self._update_ui_for_tool(tool_config, self.current_tool_option)
        self.apply_changes()

    def _update_ui_for_tool(self, tool_config: ToolConfig, tool_option: ToolOption):
        self._updating_ui = True

        self.size_scale.set_sensitive(tool_config.has_scale)
        self.color_picker.set_sensitive(tool_config.has_primary_color)
        self.extra_stack.set_visible_child_name(tool_config.shown_stack)
        self.extra_stack_revealer.set_reveal_child(tool_config.shown_stack != "empty")

        self.color_picker.set_color_list(tool_config.primary_color_list)

        for picker in (self.fill_0, self.fill_1, self.fill_2, self.outline_1, self.outline_2):
            picker.set_color_list(tool_config.secondary_color_list)

        if tool_config.has_scale:
            self.size_scale.set_value(tool_option.size)

        if tool_config.has_primary_color:
            self.color_picker.set_color(tool_option.primary_color)

        self.fill_0.set_color(tool_option.fill_color, emit=False)
        self.fill_1.set_color(tool_option.fill_color, emit=False)
        self.fill_2.set_color(tool_option.fill_color, emit=False)
        self.outline_1.set_color(tool_option.border_color, emit=False)
        self.outline_2.set_color(tool_option.border_color, emit=False)
        self.font_dropdown.set_selected_font(tool_option.font)

        self._updating_ui = False

    def _on_scroll(self, controller, dx, dy):
        active_tool_option = self._get_active_tool_option()
        if active_tool_option is None:
            return Gdk.EVENT_PROPAGATE

        modifiers = controller.get_current_event_state()
        is_zoomed = self.get_root().image_bin.get_zoom_level() != 1

        should_adjust_size = (
            (modifiers & Gdk.ModifierType.SHIFT_MASK) and (modifiers & Gdk.ModifierType.CONTROL_MASK)
        ) or not is_zoomed

        if should_adjust_size:
            adjustment = self.size_scale.get_adjustment()
            min_value = int(adjustment.get_lower())
            max_value = int(adjustment.get_upper())

            if dy == 0:
                return Gdk.EVENT_PROPAGATE

            step_unit = 0.35
            self._scroll_accum += (-step_unit if dy > 0 else step_unit)
            delta = int(self._scroll_accum)

            if delta != 0:
                self._scroll_accum -= delta
                new_size = active_tool_option.size + delta
                new_size = max(min_value, min(max_value, new_size))

                if new_size != active_tool_option.size:
                    active_tool_option.size = new_size
                    self.size_scale.set_value(new_size)
                    self.apply_changes()

            return Gdk.EVENT_STOP

        return Gdk.EVENT_PROPAGATE

    @Gtk.Template.Callback()
    def on_fill_color_changed(self, button: SimpleColorPicker, color: Gdk.RGBA):
        if self._updating_ui:
            return

        active_tool_option = self._get_active_tool_option()
        if active_tool_option is None:
            return

        active_tool_option.fill_color = color
        self.fill_0.set_color(color, emit=False)
        self.fill_1.set_color(color, emit=False)
        self.fill_2.set_color(color, emit=False)
        self.apply_changes()

    @Gtk.Template.Callback()
    def on_outline_color_changed(self, button: SimpleColorPicker, color: Gdk.RGBA):
        if self._updating_ui:
            return

        active_tool_option = self._get_active_tool_option()
        if active_tool_option is None:
            return

        active_tool_option.border_color = color
        self.outline_1.set_color(color, emit=False)
        self.outline_2.set_color(color, emit=False)
        self.apply_changes()

    @Gtk.Template.Callback()
    def on_size_scale_changed(self, adjustment: Gtk.Adjustment):
        if self._updating_ui:
            return

        active_tool_option = self._get_active_tool_option()
        if active_tool_option is None:
            return

        active_tool_option.size = int(adjustment.get_value())
        self.apply_changes()

    @Gtk.Template.Callback()
    def on_primary_color_changed(self, picker: QuickColorPicker, color: Gdk.RGBA):
        if self._updating_ui:
            return

        active_tool_option = self._get_active_tool_option()
        active_tool_config = self._get_active_tool_config()
        if active_tool_option is None or active_tool_config is None:
            return

        selected_index = picker.get_selected_index()
        if active_tool_config.match_primary_to_secondary:
            for picker in (self.fill_0, self.fill_1, self.fill_2):
                picker.set_color_by_index(selected_index)
            for picker in (self.outline_1, self.outline_2):
                picker.set_color(Gdk.RGBA(0, 0, 0, 0))

        active_tool_option.primary_color = color
        self.apply_changes()

    @Gtk.Template.Callback()
    def on_font_changed(self, dropdown: FontDropdown, font: str):
        if self._updating_ui:
            return

        active_tool_option = self._get_active_tool_option()
        if active_tool_option is None:
            return

        active_tool_option.font = font
        self.apply_changes()

    def get_current_tool(self) -> Optional[ToolOption]:
        if self.current_tool_config is None:
            return None
        return self.tool_manager.get_tool(self.current_tool_config.mode)

    def set_current_tool(self, mode: DrawingMode):
        if self.is_temp_editing:
            return

        self.drawing_tools_grid.set_current_tool(mode)
        self.current_tool_config = next(
            (tc for tc in ToolConfig.get_all_tools_positions() if tc.mode == mode), None
        )
        if self.current_tool_config:
            self.current_tool_option = self.tool_manager.get_tool(mode)
            self._update_ui_for_tool(self.current_tool_config, self.current_tool_option)
            self.apply_changes()

    def on_overlay_selected_changed(self, overlay, tool_option):
        if tool_option is not None:
            self._enter_temp_editing_mode(tool_option)
        else:
            self._exit_temp_editing_mode()

    def _enter_temp_editing_mode(self, tool_option: ToolOption):
        self.is_temp_editing = True
        self.temp_editing_tool_option = tool_option
        self.temp_editing_tool_config = next(
            (tc for tc in ToolConfig.get_all_tools_positions() if tc.mode == tool_option.mode), None
        )

        if self.temp_editing_tool_config:
            self._update_ui_for_tool(self.temp_editing_tool_config, self.temp_editing_tool_option)

        self.drawing_tools_grid.set_secondary_highlight_tool(tool_option.mode)

    def _exit_temp_editing_mode(self):
        if not self.is_temp_editing:
            return

        self.is_temp_editing = False
        self.temp_editing_tool_option = None
        self.temp_editing_tool_config = None

        self.drawing_tools_grid.set_secondary_highlight_tool(None)

        if self.current_tool_config and self.current_tool_option:
            self._update_ui_for_tool(self.current_tool_config, self.current_tool_option)



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from typing import Literal, Optional, Sequence
from dataclasses import dataclass, field
from ctypes import CDLL, POINTER, Structure, c_double, c_int, c_uint8
import json
import os

from PIL import Image

from gradia.graphics.background import Background
from gradia.utils.colors import parse_rgb_string
from gradia.constants import libdir


Step = tuple[float, str]  # (position, "rgb(r,g,b)")


class ColorStop(Structure):
    _fields_ = [
        ("position", c_double),
        ("r", c_uint8),
        ("g", c_uint8),
        ("b", c_uint8),
    ]


@dataclass
class Gradient:
    mode: Literal["linear", "conic", "radial"] = "linear"
    steps: Sequence[Step] = field(default_factory=lambda: [
        (0.0, "rgb(87,227,137)"),
        (1.0, "rgb(53,132,228)"),
    ])
    angle: float = 135.0
    name: Optional[str] = None

    def to_json(self) -> str:
        return json.dumps({
            "mode": self.mode,
            "steps": self.steps,
            "angle": self.angle,
        })

    @classmethod
    def from_json(cls, json_str: str) -> 'Gradient':
        try:
            data = json.loads(json_str)
            return cls(
                mode=data.get("mode", "linear"),
                steps=data.get("steps", [
                    (0.0, "rgb(87,227,137)"),
                    (1.0, "rgb(53,132,228)"),
                ]),
                angle=data.get("angle", 135.0),
            )
        except Exception:
            return cls()

    def to_css(self) -> str:
        steps_str = ", ".join(
            f"{color} {step * 100:.2f}%" for step, color in self.steps
        )

        if self.mode == "linear":
            return f"linear-gradient({self.angle:.2f}deg, {steps_str})"
        elif self.mode == "conic":
            return f"conic-gradient(from {self.angle:.2f}deg, {steps_str})"
        elif self.mode == "radial":
            return f"radial-gradient(circle, {steps_str})"
        else:
            return ""


@dataclass
class GradientBackground(Background):
    gradient: Gradient = field(default_factory=Gradient)
    _c_lib: Optional[CDLL] = None

    def __init__(self, gradient: Optional[Gradient] = None):
        self.gradient = gradient or Gradient()
        self._load_c_lib()

    @classmethod
    def _load_c_lib(cls) -> None:
        if cls._c_lib is not None:
            return

        gradia_path = os.path.join(libdir, "libgradient_gen.so")
        cls._c_lib = CDLL(str(gradia_path))
        cls._c_lib.generate_gradient.argtypes = [
            POINTER(c_uint8), c_int, c_int,
            POINTER(ColorStop), c_int,
            c_double, c_int
        ]
        cls._c_lib.generate_gradient.restype = None

    @classmethod
    def from_json(cls, json_str: str) -> 'GradientBackground':
        return cls(gradient=Gradient.from_json(json_str))

    def to_json(self) -> str:
        return self.gradient.to_json()

    def get_name(self) -> str:
        return "Gradient"

    def prepare_image(self, width: int, height: int) -> Image.Image:
        if self._c_lib is None:
            raise RuntimeError("C gradient library not loaded")
        return self._generate_gradient_c(width, height)

    def _generate_gradient_c(self, width: int, height: int) -> Image.Image:
        pixel_count = width * height * 4
        pixel_buffer = (c_uint8 * pixel_count)()

        steps = self.gradient.steps
        if self.gradient.mode == "conic":
            offset = (self.gradient.angle % 360.0) / 360.0
            steps = [(((pos * 0.9) + 0.05 + offset) % 1.0, color) for pos, color in steps]
            steps.sort(key=lambda s: s[0])
            steps.append((1.0, steps[0][1]))

        parsed_stops = []
        for pos, color_str in steps:
            r, g, b = parse_rgb_string(color_str)
            parsed_stops.append(ColorStop(pos, r, g, b))

        stop_array = (ColorStop * len(parsed_stops))(*parsed_stops)

        mode_map = {
            "linear": 0,
            "conic": 1,
            "radial": 2,
        }

        mode = mode_map.get(self.gradient.mode, 0)

        self._c_lib.generate_gradient(
            pixel_buffer, width, height,
            stop_array, len(parsed_stops),
            float(self.gradient.angle),
            mode
        )

        return Image.frombytes("RGBA", (width, height), bytes(pixel_buffer))




# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from gradia.graphics.gradient import Gradient

GradientPreset = tuple[str, str, int]

PREDEFINED_GRADIENTS: list[Gradient] = [
    Gradient(
        mode="linear",
        steps=[(0.0, "#f66151"), (1.0, "#ed333b")],
        angle=45,
        name=_("Red Flame")
    ),
    Gradient(
        mode="linear",
        steps=[(0.0, "#ff5f6d"), (1.0, "#ffc371")],
        angle=45,
        name=_("Sunset Orange")
    ),
    Gradient(
        mode="linear",
        steps=[(0.0, "#ffd200"), (1.0, "#f7971e")],
        angle=135,
        name=_("Golden Hour")
    ),
    Gradient(
        mode="linear",
        steps=[(0.0, "#DFFFCD"), (0.5, "#90F9C4"), (1.0, "#39F3BB")],
        angle=135,
        name=_("Mint Fresh")
    ),
    Gradient(
        mode="linear",
        steps=[(0.0, "#57e389"), (1.0, "#3584e4")],
        angle=135,
        name=_("Ocean Breeze")
    ),
    Gradient(
        mode="linear",
        steps=[(0.0, "#23d4fd"), (0.5, "#3a98f0"), (1.0, "#b721ff")],
        angle=45,
        name=_("Purple Dream")
    ),
]

PRESET_IMAGES = [
    "/be/alexandervanhee/gradia/images/adwaita.webp",
    "/be/alexandervanhee/gradia/images/blobs.webp",
    "/be/alexandervanhee/gradia/images/coral.webp",
    "/be/alexandervanhee/gradia/images/drool.webp",
    "/be/alexandervanhee/gradia/images/graph.png",
    "/be/alexandervanhee/gradia/images/mesh.webp",
    "/be/alexandervanhee/gradia/images/pink.webp",
    "/be/alexandervanhee/gradia/images/sky.webp",
    "/be/alexandervanhee/gradia/images/swirl.webp",
]

SUPPORTED_EXPORT_FORMATS = {
    'png': {
        'name': _('PNG Image (*.png)'),
        'shortname' : 'PNG',
        'mime_type': 'image/png',
        'extensions': ['.png'],
        'save_options': {'keys': [], 'values': []}
    },
    'jpeg': {
        'name': _('JPEG Image (*.jpg)'),
        'shortname' : 'JPEG',
        'mime_type': 'image/jpeg',
        'extensions': ['.jpg', '.jpeg'],
        'save_options': {'keys': ['quality'], 'values': ['90']}
    },
    'webp': {
        'name': _('WebP Image (*.webp)'),
        'shortname' : 'WebP',
        'mime_type': 'image/webp',
        'extensions': ['.webp'],
        'save_options': {'keys': ['quality'], 'values': ['90']}
    }
}

DEFAULT_EXPORT_FORMAT = 'jpeg'



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable, Optional
from gi.repository import Gdk, Gtk
from gradia.constants import rootdir

@Gtk.Template(resource_path=f"{rootdir}/ui/text_entry_popover.ui")
class TextEntryPopover(Gtk.Popover):
    __gtype_name__ = "GradiaTextEntryPopover"

    container: Gtk.Box = Gtk.Template.Child()
    scrolled: Gtk.ScrolledWindow = Gtk.Template.Child()
    text_view: Gtk.TextView = Gtk.Template.Child()
    spin: Gtk.SpinButton = Gtk.Template.Child()
    size_adjustment: Gtk.Adjustment = Gtk.Template.Child()

    def __init__(
        self,
        parent: Gtk.Widget,
        on_text_activate: Callable,
        on_text_changed: Callable,
        on_font_size_changed: Callable,
        font_size: float | int = 14,
        initial_text: Optional[str] = "",
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.set_parent(parent)

        self.original_text = initial_text
        self.original_font_size = font_size

        self.text_buffer = self.text_view.get_buffer()
        self.text_buffer.connect("changed", on_text_changed)

        key_controller = Gtk.EventControllerKey()
        key_controller.connect("key-pressed", self._on_key_pressed, on_text_activate)
        self.add_controller(key_controller)

        if initial_text:
            self.text_buffer.set_text(initial_text)
            start_iter = self.text_buffer.get_start_iter()
            end_iter = self.text_buffer.get_end_iter()
            self.text_buffer.select_range(start_iter, end_iter)

        self.size_adjustment.set_value(font_size)
        self.spin.connect("value-changed", on_font_size_changed)
        self.container.add_css_class("linked")
        self.text_view.add_css_class("rounded-top")

    def _on_key_pressed(self, controller, keyval, keycode, state, on_activate_callback):
        if keyval == Gdk.KEY_Return or keyval == Gdk.KEY_KP_Enter:
            if state & Gdk.ModifierType.CONTROL_MASK:
                on_activate_callback(self)
                return True
        elif keyval == Gdk.KEY_Escape:
            self._restore_original_values()
            self.popdown()
            return True
        return False

    def _restore_original_values(self) -> None:
        self.text_buffer.set_text(self.original_text)
        self.size_adjustment.set_value(self.original_font_size)

    def update_original_values(self, text: str, font_size: float | int) -> None:
        self.original_text = text
        self.original_font_size = font_size

    def get_text(self) -> str:
        start_iter = self.text_buffer.get_start_iter()
        end_iter = self.text_buffer.get_end_iter()
        return self.text_buffer.get_text(start_iter, end_iter, False)

    def popup_at_widget_coords(self, widget: Gtk.Widget, x: float, y: float) -> None:
        allocation = widget.get_allocation()
        rect = Gdk.Rectangle()
        rect.x = allocation.x + int(x)
        rect.y = allocation.y + int(y)
        rect.width = 1
        rect.height = 1
        self.set_pointing_to(rect)
        self.popup()
        self.text_view.grab_focus()



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any
from gi.repository import Gtk, Gdk, Graphene

class TransparencyBackground(Gtk.Widget):
    __gtype_name__ = "GradiaTransparencyBackground"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.picture_widget: Gtk.Picture | None = None
        self.square_size = 10
        self.max_tiles_x = 20
        self.max_tiles_y = 20

    def do_snapshot(self, snapshot: Gtk.Snapshot) -> None:
        offset_x, offset_y, display_width, display_height, square_size_scaled = self._calculate_geometry()

        if display_width <= 0 or display_height <= 0:
            return

        light_gray = Gdk.RGBA(red=0.9, green=0.9, blue=0.9, alpha=1.0)
        dark_gray = Gdk.RGBA(red=0.7, green=0.7, blue=0.7, alpha=1.0)

        bounds = Graphene.Rect.alloc().init(offset_x, offset_y, display_width, display_height)
        snapshot.append_color(light_gray, bounds)

        tile_size = 2 * square_size_scaled
        child_bounds = Graphene.Rect.alloc().init(offset_x, offset_y, tile_size, tile_size)

        snapshot.push_repeat(bounds, child_bounds)
        dark_square_1 = Graphene.Rect.alloc().init(offset_x + square_size_scaled, offset_y,
                                                  square_size_scaled, square_size_scaled)
        snapshot.append_color(dark_gray, dark_square_1)

        dark_square_2 = Graphene.Rect.alloc().init(offset_x, offset_y + square_size_scaled,
                                                  square_size_scaled, square_size_scaled)
        snapshot.append_color(dark_gray, dark_square_2)
        snapshot.pop()

    def set_picture_reference(self, picture: Gtk.Picture) -> None:
        self.picture_widget = picture
        if picture:
            picture.connect("notify::paintable", lambda *args: self.queue_draw())

    def _calculate_geometry(self) -> tuple[float, float, float, float, float]:
        offset_x, offset_y, display_width, display_height = self._get_image_bounds()

        scale = 1.0
        if self.picture_widget and self.picture_widget.get_paintable():
            image_width = self.picture_widget.get_paintable().get_intrinsic_width()
            image_height = self.picture_widget.get_paintable().get_intrinsic_height()
            if image_width > 0 and image_height > 0:
                scale = min(display_width / image_width, display_height / image_height)

        if scale <= 0:
            return offset_x, offset_y, display_width, display_height, 0

        max_tile_width = display_width / self.max_tiles_x
        max_tile_height = display_height / self.max_tiles_y
        square_size_scaled = max(self.square_size * scale, max_tile_width, max_tile_height)

        return offset_x, offset_y, display_width, display_height, square_size_scaled

    def _get_image_bounds(self) -> tuple[float, float, float, float]:
        if not self.picture_widget or not self.picture_widget.get_paintable():
            inset = 2
            return inset, inset, self.get_width() - 2*inset, self.get_height() - 2*inset

        widget_width = self.picture_widget.get_width()
        widget_height = self.picture_widget.get_height()
        image_width = self.picture_widget.get_paintable().get_intrinsic_width()
        image_height = self.picture_widget.get_paintable().get_intrinsic_height()

        if image_width <= 0 or image_height <= 0:
            inset = 2
            return inset, inset, widget_width - 2*inset, widget_height - 2*inset

        scale = min(widget_width / image_width, widget_height / image_height)
        display_width = image_width * scale
        display_height = image_height * scale

        offset_x = (widget_width - display_width) / 2
        offset_y = (widget_height - display_height) / 2

        inset = 2
        return offset_x + inset, offset_y + inset, display_width - 2*inset, display_height - 2*inset



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Callable, Optional
from pathlib import Path
from gi.repository import Adw, Gtk, GLib, Gdk, GdkPixbuf, Graphene, Gsk, GObject

from gradia.app_constants import PREDEFINED_GRADIENTS
from gradia.backend.settings import Settings
from gradia.constants import rootdir
from gradia.graphics.gradient import Gradient


class RecentFile:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.folder = str(path.parent)
        self.name = str(path.parent)


class RecentImageGetter:
    MAX_RESULTS = 6
    FALLBACK_PICTURES_PATH = Path.home() / "Pictures"

    def __init__(self) -> None:
        pass

    def get_recent_screenshot_files(self) -> list[RecentFile]:
        screenshots_dir = self._get_screenshots_directory()
        if not screenshots_dir or not screenshots_dir.exists():
            print(f"Screenshots directory does not exist: {screenshots_dir}")
            return None

        image_extensions = {'.png', '.jpg', '.jpeg', '.webp', '.avif'}
        all_files = [f for f in screenshots_dir.iterdir()
                     if f.is_file() and f.suffix.lower() in image_extensions]

        sorted_files = sorted(all_files, key=lambda f: f.stat().st_mtime, reverse=True)
        top_files = sorted_files[:self.MAX_RESULTS]

        return [RecentFile(f) for f in top_files]

    def _get_screenshots_directory(self) -> Path | None:
        xdg_pictures = GLib.get_user_special_dir(GLib.USER_DIRECTORY_PICTURES)
        screenshot_folder = Settings().screenshot_folder

        if screenshot_folder:
            path = Path(screenshot_folder)
            return path if path.exists() else None

        if xdg_pictures:
            path = Path(xdg_pictures) / 'Screenshots'
            return path if path.exists() else None
        return None

class ShadowedImageCard(Gtk.Widget):
    _shadows = None

    def __init__(self, path: str, radius: float = 4.0, padding: int = 8,
                 compact: bool = False):
        super().__init__()
        self.radius = radius
        self.texture = None
        self.padding = padding
        self._image_rect = Graphene.Rect()

        width = 155 if compact else 260
        height = 120 if compact else 160

        try:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(path, width, height)
            self.texture = Gdk.Texture.new_for_pixbuf(pixbuf)
        except Exception as e:
            print(f"Failed to load image {path}: {e}")

    @classmethod
    def _get_shadows(cls):
        if cls._shadows is None:
            def create_shadow(r, g, b, a, dx, dy, radius):
                color = Gdk.RGBA()
                color.red, color.green, color.blue, color.alpha = r, g, b, a
                shadow = Gsk.Shadow()
                shadow.color, shadow.dx, shadow.dy, shadow.radius = color, dx, dy, radius
                return shadow

            cls._shadows = [
                create_shadow(0, 0, 0, 0.16, 0, 0, 8),
                create_shadow(0, 0, 0, 0.24, 0, 0, 16),
                create_shadow(0, 0, 0, 0.32, 0, 0, 24)
            ]
        return cls._shadows

    def do_snapshot(self, snapshot: Gtk.Snapshot) -> None:
        if not self.texture:
            return

        widget_width = self.get_width()
        widget_height = self.get_height()
        texture_width = self.texture.get_width()
        texture_height = self.texture.get_height()

        available_width = widget_width - (self.padding * 2)
        available_height = widget_height - (self.padding * 2)

        scale_x = available_width / texture_width
        scale_y = available_height / texture_height
        scale = min(scale_x, scale_y)

        scaled_width = texture_width * scale
        scaled_height = texture_height * scale

        x_offset = self.padding + (available_width - scaled_width) / 2
        y_offset = self.padding + (available_height - scaled_height) / 2

        self._image_rect.init(x_offset, y_offset, scaled_width, scaled_height)

        snapshot.push_shadow(self._get_shadows())

        if self.radius > 0:
            rounded_rect = Gsk.RoundedRect()
            rounded_rect.init_from_rect(self._image_rect, self.radius)
            snapshot.push_rounded_clip(rounded_rect)

        snapshot.append_texture(self.texture, self._image_rect)

        if self.radius > 0:
            snapshot.pop()

        snapshot.pop()

@Gtk.Template(resource_path=f"{rootdir}/ui/recent_picker.ui")
class RecentPicker(Adw.Bin):
    __gtype_name__ = "GradiaRecentPicker"

    FRAME_SPACING = 5
    IMAGE_WIDTH = 260
    IMAGE_HEIGHT = 160
    COMPACT_IMAGE_WIDTH = 150
    COMPACT_IMAGE_HEIGHT = 120

    item_grid: Gtk.FlowBox = Gtk.Template.Child()

    recent_overlay: Gtk.Overlay= Gtk.Template.Child()
    error_overlay: Adw.StatusPage = Gtk.Template.Child()

    compact = GObject.Property(type=bool, default=False)

    def __init__(self, callback: Optional[Callable] = None, **kwargs) -> None:
        super().__init__(**kwargs)

        self.image_getter = RecentImageGetter()
        self.callback = callback

        self.image_bins: list[Gtk.Button] = []
        self.recent_files: list[RecentFile] = []

        self.gradient_colors = PREDEFINED_GRADIENTS
        self.original_gradient_indexes = list(range(len(self.gradient_colors)))
        combined = list(zip(self.gradient_colors, self.original_gradient_indexes))
        self.gradient_colors, self.original_gradient_indexes = zip(*combined)
        self.gradient_colors = list(self.gradient_colors)
        self.original_gradient_indexes = list(self.original_gradient_indexes)

        self.connect("notify::compact", self._on_compact_changed)
        self._setup_cards()
        self._load_images()

    def _on_compact_changed(self, *args) -> None:
        self._setup_cards()
        self._load_images()

    def _setup_cards(self) -> None:
        for child in list(self.item_grid):
            self.item_grid.remove(child)

        self.image_bins.clear()

        width = self.COMPACT_IMAGE_WIDTH if self.compact else self.IMAGE_WIDTH
        height = self.COMPACT_IMAGE_HEIGHT if self.compact else self.IMAGE_HEIGHT

        for index in range(6):
            image_bin = Adw.Bin()
            image_bin.set_size_request(width, height)
            image_bin.add_css_class("card")
            self._apply_gradient_to_button(image_bin, index)
            placeholder = Gtk.Box()
            image_bin.set_child(placeholder)
            self.image_bins.append(image_bin)
            self.item_grid.append(image_bin)

        self.item_grid.connect(
            "child-activated",
            lambda flowbox, flowbox_child: self._on_image_clicked(
                self.image_bins.index(flowbox_child.get_child())
            )
        )

    def _on_image_clicked(self, index: int, *args) -> None:
        if index < len(self.recent_files):
            file_path = self.recent_files[index].path
            original_gradient_index = self.original_gradient_indexes[index % len(self.original_gradient_indexes)]

            if self.callback:
                self.callback(str(file_path), original_gradient_index)

    def refresh(self) -> None:
        self._load_images()

    def _apply_gradient_to_button(self, button: Gtk.Button, index: int) -> None:
        gradient_name = f"gradient-button-{index}"
        button.set_name(gradient_name)
        button.add_css_class("recent-button")

        color_index = index % len(self.gradient_colors)
        gradient = self.gradient_colors[color_index]

        css = f"""
            *#{gradient_name} {{
                background-image: {gradient.to_css()};
            }}
        """

        css_provider = Gtk.CssProvider()
        css_provider.load_from_string(css)
        button.get_style_context().add_provider(
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def _load_images(self) -> None:
        recent_files = self.image_getter.get_recent_screenshot_files()
        self._update_display(recent_files)

    def _update_display(self, recent_files: list[RecentFile]) -> None:
        self.recent_files = recent_files

        if self.recent_files is None:
            self.error_overlay.set_visible(True)
            self.item_grid.set_opacity(0.25)
            recent_files = []
        else:
            self.error_overlay.set_visible(False)
            self.item_grid.set_opacity(1)

        radius = 2.0 if self.compact else 4.0
        padding = 10

        for i in range(6):
            if i < len(recent_files):
                file = recent_files[i]

                try:
                    shadowed = ShadowedImageCard(str(file.path), radius=radius, padding=padding, compact=self.compact)
                    self.image_bins[i].set_child(shadowed)
                    self.image_bins[i].set_sensitive(True)

                    label = file.path.name
                    self.image_bins[i].set_tooltip_text(label)

                    flowbox_child = self.image_bins[i].get_parent()
                    if flowbox_child:
                        flowbox_child.update_property(
                            [Gtk.AccessibleProperty.LABEL],
                            [label]
                        )
                except Exception as e:
                    icon = Gtk.Image.new_from_icon_name("image-missing-symbolic")
                    self.image_bins[i].set_child(icon)
                    self.image_bins[i].set_sensitive(False)
                    self.image_bins[i].set_tooltip_text("")
                    print(f"Error loading image {file.path}: {e}")
            else:
                icon = Gtk.Image.new_from_icon_name("image-missing-symbolic")
                self.image_bins[i].set_child(icon)
                self.image_bins[i].set_sensitive(False)
                self.image_bins[i].set_tooltip_text("")



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import pytesseract
import os
import json
import locale
import gi
from dataclasses import dataclass
from typing import List, ClassVar, Optional
gi.require_version("Soup", "3.0")
from gi.repository import Soup, GLib, Gio
from pathlib import Path
from gradia.backend.logger import Logger
from gradia.backend.settings import Settings
from gradia.constants import app_id, ocr_enabled
from gradia.constants import ocr_tesseract_cmd, ocr_original_tessdata


logger = Logger()

@dataclass(frozen=True)
class OCRModel:
    code: str
    name: str
    size: int

class OCR:
    _tessdata_cache: ClassVar[Optional[dict]] = None

    def __init__(self, window=None):
        self.tesseract_cmd = ocr_tesseract_cmd
        self.original_tessdata_dir = ocr_original_tessdata
        self.user_tessdata_dir = os.path.join(GLib.get_user_data_dir(), "tessdata")
        self.window = window

        pytesseract.pytesseract.tesseract_cmd = self.tesseract_cmd
        self._session = None
        self.settings = Settings()

        self._update_ocr_action_state()

    def _load_tessdata_json(self) -> dict:
        if OCR._tessdata_cache is not None:
            return OCR._tessdata_cache
        data = Gio.resources_lookup_data(
            "/be/alexandervanhee/gradia/models.json",
            Gio.ResourceLookupFlags.NONE
        )
        OCR._tessdata_cache = json.loads(data.get_data().decode("utf-8"))
        return OCR._tessdata_cache

    def _get_ui_language(self) -> Optional[str]:
        for var in ("LANGUAGE", "LC_ALL", "LC_MESSAGES", "LANG"):
            val = os.environ.get(var)
            if val:
                lang = val.split(":")[0].split(".")[0]
                if lang and lang != "C" and lang != "POSIX":
                    return lang
        try:
            lang, _ = locale.getlocale()
            if lang:
                return lang
        except Exception:
            pass
        return None

    def _get_translated_name(self, code: str, entry: dict) -> str:
        ui_lang = self._get_ui_language()
        if ui_lang:
            translations = entry.get("translations", {})
            name = translations.get(ui_lang)
            if not name and "_" in ui_lang:
                name = translations.get(ui_lang.split("_")[0])
            if name:
                return name
        return entry.get("english_name", code)

    def get_downloadable_models(self) -> List[OCRModel]:
        data = self._load_tessdata_json()
        models = []
        for code, entry in data.items():
            name = self._get_translated_name(code, entry)
            size = entry.get("size_bytes", 0)
            models.append(OCRModel(code=code, name=name, size=size))
        return models

    def _update_ocr_action_state(self):
        if self.window:
            available = Path(ocr_tesseract_cmd).exists()
            ocr_action = self.window.lookup_action("ocr")
            if ocr_action and hasattr(ocr_action, 'set_state'):
                ocr_action.set_state(GLib.Variant.new_boolean(available and ocr_enabled.lower() == 'true'))

    def get_current_model(self):
        return self.settings.trained_data

    def set_current_model(self, model_code: str):
        if self.is_model_installed(model_code):
            self.settings.trained_data = model_code
            logger.info(f"Setting current OCR model to: {model_code}")
        else:
            logger.warning(f"Cannot set model {model_code}: not installed")
            raise ValueError(f"Model {model_code} is not installed")

    def extract_text(self, image, primary_lang):
        if not self.get_installed_models():
            raise RuntimeError("No OCR language models are available")

        if not self.is_model_installed(primary_lang):
            raise RuntimeError(f"OCR language model '{primary_lang}' is not installed")

        self.set_current_model(primary_lang)

        try:
            tessdata_dir = self._get_tessdata_dir_for_lang(primary_lang)
            config = f'--tessdata-dir "{tessdata_dir}"'
            lang = primary_lang
            if self.is_model_installed("eng") and primary_lang != "eng":
                lang = f"{primary_lang}+eng"

            extracted_text = pytesseract.image_to_string(
                image,
                lang=lang,
                config=config
            )
            return extracted_text.strip()
        except FileNotFoundError:
            raise
        except Exception as e:
            raise Exception(f"OCR processing failed: {str(e)}")

    def _get_tessdata_dir_for_lang(self, lang_code):
        user_model_path = Path(self.user_tessdata_dir) / f"{lang_code}.traineddata"
        if user_model_path.exists():
            return self.user_tessdata_dir
        return self.original_tessdata_dir

    def get_installed_models(self):
        installed = set()

        for path in [self.original_tessdata_dir, self.user_tessdata_dir]:
            p = Path(path)
            if p.exists():
                for file in p.glob("*.traineddata"):
                    if file.stem != "osd":
                        installed.add(file.stem)

        return sorted(list(installed))

    def is_model_installed(self, model_code: str):
        return model_code in self.get_installed_models()

    def download_model(self, model_code: str, progress_callback=None):
        if not self._session:
            self._session = Soup.Session()

        url = f"https://github.com/tesseract-ocr/tessdata_best/raw/4.1.0/{model_code}.traineddata"
        output_path = Path(self.user_tessdata_dir) / f"{model_code}.traineddata"

        output_path.parent.mkdir(parents=True, exist_ok=True)

        message = Soup.Message.new("GET", url)

        def on_download_complete(session, result, user_data):
            try:
                glib_bytes = session.send_and_read_finish(result)
                if message.get_status() != Soup.Status.OK:
                    raise RuntimeError(f"HTTP error {message.get_status()}")

                raw_bytes = glib_bytes.get_data()

                with open(output_path, 'wb') as f:
                    f.write(raw_bytes)
                    logger.info(f"saving to {output_path}")

                logger.info(f"Downloaded OCR model: {model_code}")
                self.set_current_model(model_code)
                self._update_ocr_action_state()

                if progress_callback:
                    GLib.idle_add(progress_callback, True, f"Downloaded {model_code}")

            except Exception as e:
                logger.error(f"Failed to download OCR model {model_code}: {e}")
                if progress_callback:
                    GLib.idle_add(progress_callback, False, str(e))

        self._session.send_and_read_async(
            message,
            GLib.PRIORITY_DEFAULT,
            None,
            on_download_complete,
            None
        )

    def delete_model(self, model_code: str):
        user_model_path = Path(self.user_tessdata_dir) / f"{model_code}.traineddata"

        if user_model_path.exists():
            user_model_path.unlink()
            logger.info(f"Deleted OCR model: {model_code}")

            if self.get_current_model() == model_code:
                available_models = self.get_installed_models()
                if available_models:
                    self.set_current_model(available_models[0])

            self._update_ocr_action_state()
            return True
        else:
            logger.warning(f"OCR model not found: {model_code}")
            return False



import sys
import os
import shutil
from pathlib import Path

from PIL import Image
from io import BytesIO

from typing import Optional
from gradia.backend.logger import Logger
from gradia.utils.timestamp_filename import TimestampedFilenameGenerator
logging = Logger()

class StdinImageLoader:
    def __init__(self):
        self.temp_path: Optional[str] = None

    def get_flatpak_safe_temp_dir(self) -> str:
        xdg_cache_home = os.environ.get("XDG_CACHE_HOME") or os.path.expanduser("~/.cache")
        temp_dir = os.path.join(xdg_cache_home, "gradia", "stdin")
        os.makedirs(temp_dir, exist_ok=True)
        return temp_dir

    def read_from_stdin(self) -> Optional[str]:
        if sys.stdin.isatty():
            return None

        try:
            logging.debug("Reading image from stdin…")
            image_data = sys.stdin.buffer.read()
            if not image_data:
                raise ValueError("No image data received from stdin.")

            image = Image.open(BytesIO(image_data))
            image.load()

            temp_dir = self.get_flatpak_safe_temp_dir()
            filename = TimestampedFilenameGenerator().generate(_("Edited Image From %Y-%m-%d %H-%M-%S")) + ".png"
            temp_path = os.path.join(temp_dir, filename)
            Path(temp_dir).mkdir(parents=True, exist_ok=True)

            image.save(temp_path)
            self.temp_path = temp_path

            logging.info(f"Temporary image file written to: {self.temp_path}")

            return self.temp_path

        except Exception as e:
            logging.critical("Failed to read image from stdin.", exception=e, show_exception=True)
            return None



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk, GObject
from typing import Optional

class ToggleGroup(Gtk.Box):
    __gtype_name__ = "GradiaToggleGroup"

    active_name = GObject.Property(
        type=str,
        default="",
        flags=GObject.ParamFlags.READWRITE
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self._updating = False
        self._active_name_value = ""
        self._toggle_buttons = {}
        self._create_widget()

    def _create_widget(self) -> None:
        flowbox = Gtk.FlowBox()
        flowbox.set_max_children_per_line(4)
        flowbox.set_min_children_per_line(2)
        flowbox.set_homogeneous(True)
        flowbox.set_row_spacing(10)
        flowbox.set_column_spacing(8)

        toggle_data = [
            ("none", _("_None")),
            ("solid", _("_Solid")),
            ("gradient", _("_Gradient")),
            ("image", _("_Image"))
        ]

        for name, label in toggle_data:
            toggle = Gtk.ToggleButton()
            toggle.set_label(label)
            toggle.set_use_underline(True)
            toggle.connect("toggled", self._on_toggle_changed, name)
            self._toggle_buttons[name] = toggle
            flowbox.append(toggle)

        self.append(flowbox)

    def _on_toggle_changed(self, button: Gtk.ToggleButton, name: str) -> None:
        if self._updating:
            return

        if button.get_active():
            self._updating = True
            for toggle_name, toggle_button in self._toggle_buttons.items():
                if toggle_name != name:
                    toggle_button.set_active(False)
            self._updating = False

            self._active_name_value = name
            self.notify("active-name")
        else:
            if self._active_name_value == name:
                self._updating = True
                button.set_active(True)
                self._updating = False

    def do_get_property(self, pspec):
        if pspec.name == "active-name":
            return self._active_name_value

    def get_active_name(self) -> Optional[str]:
        return self._active_name_value if self._active_name_value else None

    def set_active_name(self, name: str) -> None:
        if name in self._toggle_buttons and self._active_name_value != name:
            self._active_name_value = name
            self._updating = True
            for toggle_name, toggle_button in self._toggle_buttons.items():
                toggle_button.set_active(toggle_name == name)
            self._updating = False
            self.notify("active-name")



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Optional, Callable
from gi.repository import Gtk, GObject
from gradia.overlay.drawing_actions import DrawingMode
from gradia.backend.settings import Settings
from gradia.backend.tool_config import ToolConfig
settings = Settings()

class DrawingToolsGrid(Gtk.Grid):
    __gtype_name__ = "GradiaDrawingToolsGrid"

    __gsignals__ = {
        "tool-changed": (GObject.SignalFlags.RUN_FIRST, None, (object,))
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_row_spacing(9)
        self.set_column_spacing(18)
        self.set_column_homogeneous(True)
        self.set_row_homogeneous(True)

        self._buttons = {}
        self._current_tool = None
        self._tool_configs = {}

        self._create_tool_buttons()

        self.connect("realize", self._on_realize)

    def _on_realize(self, widget):
        initial_mode = DrawingMode(settings.draw_mode)
        self._select_tool(initial_mode)
        if initial_mode in self._tool_configs:
            self._current_tool = initial_mode
            self.emit("tool-changed", self._tool_configs[initial_mode])

        widget.disconnect_by_func(self._on_realize)

    def _create_tool_buttons(self):
        tools = ToolConfig.get_all_tools_positions()

        for tool_config in tools:
            button = Gtk.ToggleButton(
                icon_name=tool_config.icon,
                tooltip_text=tool_config.mode.label(),
                width_request=40,
                height_request=40,
                css_classes=["flat", "circular"]
            )

            button.get_first_child().set_pixel_size(18)

            self._tool_configs[tool_config.mode] = tool_config
            self._buttons[tool_config.mode] = button

            button.connect("toggled", self._on_button_toggled, tool_config.mode)

            self.attach(button, tool_config.column, tool_config.row, 1, 1)

    def _on_button_toggled(self, button: Gtk.ToggleButton, mode: DrawingMode):
        if button.get_active():
            for other_mode, other_button in self._buttons.items():
                if other_mode != mode and other_button.get_active():
                    other_button.handler_block_by_func(self._on_button_toggled)
                    other_button.set_active(False)
                    other_button.handler_unblock_by_func(self._on_button_toggled)

            old_tool = self._current_tool
            self._current_tool = mode

            settings.draw_mode = mode.value

            if old_tool != mode:
                tool_config = self._tool_configs[mode]
                self.emit("tool-changed", tool_config)
        else:
            if self._current_tool == mode:
                button.handler_block_by_func(self._on_button_toggled)
                button.set_active(True)
                button.handler_unblock_by_func(self._on_button_toggled)

    def _select_tool(self, mode: DrawingMode):
        if mode in self._buttons:
            self._buttons[mode].set_active(True)

    def get_current_tool(self) -> Optional[ToolConfig]:
        if self._current_tool:
            return self._tool_configs[self._current_tool]
        return None

    def set_current_tool(self, mode: DrawingMode):
        self._select_tool(mode)

    def set_secondary_highlight_tool(self, mode: Optional[DrawingMode]):
        for button in self._buttons.values():
            button.add_css_class("flat")
            button.remove_css_class("accent")
        if mode and mode in self._buttons:
            self._buttons[mode].remove_css_class("flat")
            self._buttons[mode].add_css_class("accent")



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections.abc import Callable
from typing import Optional

from gi.repository import GObject, Gtk, Adw, GLib

from gradia.graphics.gradient import GradientBackground
from gradia.graphics.gradient_selector import GradientSelector
from gradia.graphics.solid import SolidSelector, SolidBackground
from gradia.graphics.image import ImageSelector, ImageBackground
from gradia.graphics.background import Background
from gradia.ui.widget.toggle_group import ToggleGroup
from gradia.constants import rootdir  # pyright: ignore
from gradia.backend.settings import Settings


MODES = ["none", "solid", "gradient", "image"]

@Gtk.Template(resource_path=f"{rootdir}/ui/background_selector.ui")
class BackgroundSelector(Adw.Bin):
    __gtype_name__ = "GradiaBackgroundSelector"

    toggle_group: ToggleGroup = Gtk.Template.Child()
    stack: Gtk.Stack = Gtk.Template.Child()
    stack_revealer: Gtk.Revealer = Gtk.Template.Child()

    def __init__(
        self,
        callback: Optional[Callable[[Background], None]] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)

        self.settings = Settings()
        self.solid = SolidBackground.from_json(self.settings.solid_state or '{}')
        self.gradient = GradientBackground.from_json(self.settings.gradient_state or '{}')
        self.image = ImageBackground()
        self.callback = callback
        self.current_mode_callback = None
        self.current_mode = self.settings.background_mode if self.settings.background_mode in MODES else "gradient"
        self.initial_mode = self.current_mode

        self.gradient_selector = GradientSelector(self.gradient, self._on_gradient_changed)
        self.solid_selector = SolidSelector(self.solid, self._on_solid_changed)
        self.image_selector = ImageSelector(self.image, self._on_image_changed)

        self._setup()

    """
    Setup Methods
    """

    def _setup(self) -> None:
        self.toggle_group.set_active_name(self.current_mode)

        self.stack.add_named(self.solid_selector, "solid")
        self.stack.add_named(self.gradient_selector, "gradient")
        self.stack.add_named(self.image_selector, "image")
        if self.current_mode != "none":
            self.stack.set_visible_child_name(self.current_mode)
        self._update_revealer_visibility()

    """
    Callbacks
    """

    @Gtk.Template.Callback()
    def _on_group_changed(self, group: Adw.ToggleGroup, _param: GObject.ParamSpec, *args) -> None:
        active_name = group.get_active_name()
        if active_name in MODES and active_name != self.current_mode:
            self.current_mode = active_name
            self.settings.background_mode = active_name
            if self.current_mode != "none":
                self.stack.set_visible_child_name(active_name)
            self._update_revealer_visibility()
            self._notify_current()

    def _on_gradient_changed(self, gradient: GradientBackground) -> None:
        self.settings.gradient_state = gradient.to_json()
        if self.current_mode == "gradient":
            self._notify_current()

    def _on_solid_changed(self, solid: SolidBackground) -> None:
        self.settings.solid_state = solid.to_json()
        if self.current_mode == "solid":
            self._notify_current()

    def _on_image_changed(self, _image: ImageBackground) -> None:
        if self.current_mode == "image":
            self._notify_current()

    def set_current_mode_callback(self, callback: Callable[[str], None]) -> None:
        self.current_mode_callback = callback
        self._notify_current()

    """
    Internal Methods
    """

    def _update_revealer_visibility(self) -> None:
        should_reveal = self.current_mode != "none"
        currently_revealed = self.stack_revealer.get_child_revealed()
        self.stack_revealer.set_reveal_child(should_reveal)

        if should_reveal:
            if not currently_revealed:
                GLib.timeout_add(300, lambda: (
                    self.stack_revealer.set_overflow(Gtk.Overflow.VISIBLE), False
                )[1])
            else:
                self.stack_revealer.set_overflow(Gtk.Overflow.VISIBLE)
        else:
            self.stack_revealer.set_overflow(Gtk.Overflow.HIDDEN)

    # TODO: Fix callback type error
    def _notify_current(self) -> None:
        if self.callback:
            current_background = self.get_current_background()
            self.callback(current_background)
        if self.current_mode_callback:
            self.current_mode_callback(self.current_mode)

    def get_current_background(self) -> GradientBackground | SolidBackground | ImageBackground | None:
        backgrounds: dict[str, GradientBackground | SolidBackground | ImageBackground] = {
            "gradient": self.gradient,
            "solid": self.solid,
            "image": self.image
        }
        return backgrounds.get(self.current_mode)



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
from pathlib import Path
from gi.repository import Gtk, Adw, GLib, Gio, GObject
from typing import Optional

from gradia.constants import rootdir, ocr_enabled  # pyright: ignore
from gradia.backend.settings import Settings
from gradia.app_constants import SUPPORTED_EXPORT_FORMATS
from gradia.backend.logger import Logger
from gradia.ui.preferences.provider_selection_window import ProviderListPage
from gradia.ui.preferences.ocr_model_page import OCRModelPage
from gradia.backend.ocr import OCR

logger = Logger()


@Gtk.Template(resource_path=f"{rootdir}/ui/preferences_window.ui")
class PreferencesWindow(Adw.PreferencesDialog):
    __gtype_name__ = 'GradiaPreferencesWindow'

    help_button: Gtk.Button = Gtk.Template.Child()
    save_format_group: Adw.PreferencesGroup = Gtk.Template.Child()
    delete_screenshot_switch: Adw.SwitchRow = Gtk.Template.Child()
    overwrite_screenshot_switch: Adw.SwitchRow = Gtk.Template.Child()
    confirm_upload_switch: Adw.SwitchRow = Gtk.Template.Child()
    save_format_combo: Adw.ComboRow = Gtk.Template.Child()
    provider_name: Gtk.Label = Gtk.Template.Child()
    exiting_combo: Adw.ComboRow = Gtk.Template.Child()
    folder_label: Gtk.Label = Gtk.Template.Child()

    ocr_enabled = GObject.Property(type=bool, default=ocr_enabled.lower() == 'true')

    def __init__(self, parent_window: Adw.ApplicationWindow, **kwargs):
        super().__init__(**kwargs)

        self.parent_window = parent_window

        self.settings = Settings()

        self._setup_widgets()
        self._connect_signals()

        shortcut_controller = Gtk.ShortcutController()
        shortcut = Gtk.Shortcut.new(
            Gtk.ShortcutTrigger.parse_string("Escape"),
            Gtk.ShortcutAction.parse_string("action(window.close)")
        )
        shortcut_controller.add_shortcut(shortcut)
        self.add_controller(shortcut_controller)

    def _setup_widgets(self):
        self._setup_save_format_combo()
        self._setup_exiting_combo()
        self._setup_provider_display()
        self._bind_settings()
        self._setup_folder_label()

    def _setup_provider_display(self):
        provider_name = self.settings.provider_name
        if provider_name:
            self.provider_name.set_text(provider_name)
        else:
            self.provider_name.set_text(_("None Selected"))

    def _setup_folder_label(self):
       path = self.settings.screenshot_folder
       if path is None:
           path = "Screenshots"
       self.folder_label.set_text(os.path.basename(path))

    def _setup_save_format_combo(self):
        current_format = self.settings.export_format
        string_list = Gtk.StringList()

        format_keys = list(SUPPORTED_EXPORT_FORMATS.keys())
        self.format_keys = format_keys

        for fmt in format_keys:
            display_name = SUPPORTED_EXPORT_FORMATS[fmt]['shortname']
            string_list.append(display_name)

        self.save_format_combo.set_model(string_list)

        try:
            current_index = format_keys.index(current_format)
            self.save_format_combo.set_selected(current_index)
        except ValueError:
            self.save_format_combo.set_selected(0)

        self.save_format_combo.connect("notify::selected", self._on_save_format_changed)

    def _setup_exiting_combo(self):
        current_exit_method = self.settings.exit_method
        string_list = Gtk.StringList()

        exit_options = [
            ("confirm", _("Ask for Confirmation")),
            ("copy", _("Copy and Close ")),
            ("none", _("Close Instantly"))
        ]
        self.exit_option_keys = [key for key, _ in exit_options]

        for key, display_name in exit_options:
            string_list.append(display_name)

        self.exiting_combo.set_model(string_list)

        try:
            current_index = self.exit_option_keys.index(current_exit_method)
            self.exiting_combo.set_selected(current_index)
        except ValueError:
            self.exiting_combo.set_selected(0)

        self.exiting_combo.connect("notify::selected", self._on_exit_method_changed)

    def _on_save_format_changed(self, combo_row, pspec) -> None:
        selected = combo_row.get_selected()
        if selected < len(self.format_keys):
            self.settings.export_format = self.format_keys[selected]

    def _on_exit_method_changed(self, combo_row, pspec) -> None:
        selected = combo_row.get_selected()
        if selected < len(self.exit_option_keys):
            self.settings.exit_method = self.exit_option_keys[selected]

    def _connect_signals(self):
        self.help_button.connect("activated", self._on_help_button_clicked)

    def _on_help_button_clicked(self, button: Gtk.Button) -> None:
        self.push_subpage(ScreenshotGuidePage(self))

    def _copy_to_clipboard(self, text: str) -> None:
        clipboard = self.get_clipboard()
        clipboard.set(text)
        self.show_toast(_("Copied!"))

    def show_toast(self, message: str) -> None:
        toast = Adw.Toast.new(message)
        toast.set_timeout(2)
        if hasattr(self.parent_window, 'add_toast'):
            self.parent_window.add_toast(toast)

    def _bind_settings(self):
        self.settings.bind_switch(self.delete_screenshot_switch,"trash-screenshots-on-close")
        self.settings.bind_switch(self.confirm_upload_switch,"show-export-confirm-dialog")
        self.settings.bind_switch(self.overwrite_screenshot_switch,"overwrite-screenshot")

    @Gtk.Template.Callback()
    def on_choose_provider_clicked(self, button: Gtk.Button) -> None:
        def handle_selection(name: str, command: str):
            self.provider_name.set_text(name)
            self.settings.provider_name = name
            self.settings.custom_export_command = command
            self.parent_window.update_command_ready()
        self.push_subpage(ProviderListPage(preferences_dialog=self,on_provider_selected=handle_selection))

    @Gtk.Template.Callback()
    def on_folder_row_clicked(self, row: Adw.ActionRow) -> None:
        file_dialog = Gtk.FileDialog()
        def on_folder_selected(dialog, result):
            try:
                folder = dialog.select_folder_finish(result)
                if folder:
                    folder_path = folder.get_path()
                    logger.info(f"Selected folder for home page: {folder_path}")
                    self.folder_label.set_text(os.path.basename(folder_path))
                    self.settings.screenshot_folder = folder_path

                    window = self.parent_window
                    action = window.lookup_action("set-screenshot-folder") if window else None
                    if action:
                        action.activate(GLib.Variant('s', folder_path))

            except GLib.Error:
                pass

        file_dialog.select_folder(
            parent=self.get_root(),
            cancellable=None,
            callback=on_folder_selected
        )

    @Gtk.Template.Callback()
    def on_manage_language_models_clicked(self, row: Adw.ActionRow) -> None:
        self.push_subpage(OCRModelPage(preferences_dialog=self, window=self.get_root()))

@Gtk.Template(resource_path=f"{rootdir}/ui/preferences/screenshot_guide_page.ui")
class ScreenshotGuidePage(Adw.NavigationPage):
    __gtype_name__ = 'GradiaScreenshotGuidePage'

    interactive_entry: Gtk.Entry = Gtk.Template.Child()
    fullscreen_entry: Gtk.Entry = Gtk.Template.Child()
    delayed_entry: Gtk.Entry = Gtk.Template.Child()

    def __init__(self, preferences_dialog, **kwargs):
        super().__init__(**kwargs)
        self.preferences_dialog = preferences_dialog
        self._connect_signals()
        self._setup_command_entries()

    def _is_running_in_flatpak(self) -> bool:
        if os.getenv('FLATPAK_ID'):
            return True
        if Path('/.flatpak-info').exists():
            return True
        if '/app/' in str(Path(__file__).resolve()):
            return True
        return False

    def _get_command_for_screenshot_type(self, screenshot_type: str) -> str:
        if self._is_running_in_flatpak():
            return f"flatpak run be.alexandervanhee.gradia --screenshot={screenshot_type}"
        else:
            return f"gradia --screenshot={screenshot_type}"

    def _connect_signals(self):
        self.interactive_entry.connect("icon-press", self._on_entry_icon_press)
        self.fullscreen_entry.connect("icon-press", self._on_entry_icon_press)
        self.delayed_entry.connect("icon-press", self._on_entry_icon_press)

    def _on_entry_icon_press(self, entry: Gtk.Entry, icon_pos: Gtk.EntryIconPosition) -> None:
        if icon_pos == Gtk.EntryIconPosition.SECONDARY:
            self._copy_to_clipboard(entry.get_text())

    def _copy_to_clipboard(self, text: str) -> None:
        clipboard = self.get_clipboard()
        clipboard.set(text)
        toast = Adw.Toast(title=_("Copied!"), timeout=1.5)
        self.preferences_dialog.add_toast(toast)

    def _setup_command_entries(self):
        interactive_command = self._get_command_for_screenshot_type("INTERACTIVE")
        fullscreen_command = self._get_command_for_screenshot_type("FULL")

        self.interactive_entry.set_text(interactive_command)
        self.fullscreen_entry.set_text(fullscreen_command)



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk, Adw, Gdk, GObject, Gio
from gradia.utils.colors import is_light_color_rgba, rgba_to_hex

class ColorPickerMixin:
    def _get_base_colors(self, alpha=1.0, secondary=False):
        base_colors = [
            (Gdk.RGBA(0.88, 0.11, 0.14, alpha), _("Red")),
            (Gdk.RGBA(0.18, 0.76, 0.49, alpha), _("Green")),
            (Gdk.RGBA(0.21, 0.52, 0.89, alpha), _("Blue")),
            (Gdk.RGBA(0.96, 0.83, 0.18, alpha), _("Yellow")),
            (Gdk.RGBA(0.0, 0.0, 0.0, alpha), _("Black")),
            (Gdk.RGBA(1.0, 1.0, 1.0, alpha), _("White")),
            ]

        if secondary:
            base_colors.append((Gdk.RGBA(0, 0, 0, 0), _("Transparent")))


        return base_colors

    def _create_color_button(self, color, name):
        button = Gtk.Button()
        button.set_has_frame(False)
        button.add_css_class('flat')
        button.add_css_class('color-hover-bg')

        color_box = Gtk.Box()
        color_box.add_css_class('color-button')
        color_box.set_size_request(24, 24)
        button.set_child(color_box)

        self._apply_color_to_box(color_box, color)
        self._apply_hover_background(button, color)

        button.set_tooltip_text(name)
        return button

    def _apply_color_to_box(self, box, color):
        ctx = box.get_style_context()
        if hasattr(box, "_color_css_provider"):
            ctx.remove_provider(box._color_css_provider)

        if color.alpha == 0:
            css = ".color-button { background-color: #b2b2b2; }"
        elif color.alpha < 1.0:
            red = int((color.red * color.alpha + 1.0 * (1.0 - color.alpha)) * 255)
            green = int((color.green * color.alpha + 1.0 * (1.0 - color.alpha)) * 255)
            blue = int((color.blue * color.alpha + 1.0 * (1.0 - color.alpha)) * 255)
            css = f".color-button {{ background-color: rgb({red}, {green}, {blue}); }}"
        else:
            css = f".color-button {{ background-color: rgb({int(color.red * 255)}, {int(color.green * 255)}, {int(color.blue * 255)}); }}"

        provider = Gtk.CssProvider()
        provider.load_from_data(css.encode())
        ctx.add_provider(provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        box._color_css_provider = provider
        ctx.remove_class("transparent-color-button-small") if color.alpha > 0 else ctx.add_class("transparent-color-button-small")

    def _apply_hover_background(self, widget, color):
        if color.alpha == 0.0:
            rgba_str = "rgba(128, 128, 128, 0.15)"
        else:
            rgba_str = f"rgba({int(color.red * 255)}, {int(color.green * 255)}, {int(color.blue * 255)}, {color.alpha * 0.15})"

        css_provider = Gtk.CssProvider()
        css = f"""
        .color-hover-bg:hover {{
            background-color: {rgba_str};
        }}
        """
        css_provider.load_from_data(css.encode())
        widget.get_style_context().add_provider(
            css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def _create_more_colors_button(self):
        more_colors_button = Gtk.Button()
        more_colors_button.set_has_frame(False)
        more_colors_button.add_css_class('flat')
        more_colors_button.add_css_class('color-hover-bg')
        more_colors_icon = Gtk.Image.new_from_icon_name('color-symbolic')
        more_colors_icon.set_icon_size(Gtk.IconSize.NORMAL)
        more_colors_button.set_child(more_colors_icon)
        more_colors_button.set_tooltip_text(_('More Colors'))
        return more_colors_button

    def _show_color_dialog(self, callback):
        color_dialog = Gtk.ColorDialog()
        color_dialog.set_title(_("Choose Color"))
        color_dialog.set_with_alpha(self.with_alpha)
        toplevel = None
        color_dialog.choose_rgba(
            toplevel,
            self.get_property('color'),
            None,
            callback
        )
    def get_selected_index(self):
        if not self._selected_button:
            return None
        return list(self).index(self._selected_button)

class QuickColorPicker(Gtk.Box, ColorPickerMixin):
    __gtype_name__ = 'GradiaQuickColorPicker'

    color = GObject.Property(
        type=Gdk.RGBA,
        default=Gdk.RGBA(0.2, 0.4, 1.0, 1.0),
        flags=GObject.ParamFlags.READWRITE
    )

    with_alpha = GObject.Property(
        type=bool,
        default=True,
        flags=GObject.ParamFlags.READWRITE
    )

    __gsignals__ = {
        'color-changed': (GObject.SignalFlags.RUN_FIRST, None, (Gdk.RGBA,))
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_orientation(Gtk.Orientation.HORIZONTAL)
        self.set_halign(Gtk.Align.CENTER)
        self.set_spacing(4)
        self._selected_button = None
        self._custom_colors = None
        self._setup_ui()
        self._setup_bindings()

    def _setup_ui(self):
        self._create_color_row()

    def _create_color_row(self):
        while self.get_first_child():
            self.remove(self.get_first_child())
        self._selected_button = None

        if self._custom_colors is not None:
            self.color_palette = self._custom_colors
        else:
            self.color_palette = self._get_base_colors()

        for color, name in self.color_palette:
            color_button = self._create_color_button(color, name or rgba_to_hex(color))
            checkmark = Gtk.Image.new_from_icon_name("object-select-symbolic")
            checkmark.set_pixel_size(12)
            checkmark.add_css_class("checkmark-icon")
            if is_light_color_rgba(color):
                checkmark.add_css_class("dark")

            color_box = color_button.get_child()
            color_button.set_child(None)

            overlay = Gtk.Overlay(width_request=20, height_request=20)
            overlay.add_overlay(checkmark)
            overlay.set_halign(Gtk.Align.CENTER)
            overlay.set_valign(Gtk.Align.CENTER)
            overlay.set_child(color_box)
            color_button.set_child(overlay)

            color_button._color = color
            color_button._checkmark = checkmark
            color_button.connect('clicked', lambda btn, c=color: self._on_color_selected(c, btn))
            self.append(color_button)

        more_colors_button = self._create_more_colors_button()
        more_colors_button.connect('clicked', self._on_more_colors_clicked)
        self.append(more_colors_button)
        self._update_selection()

    def _setup_bindings(self):
        self.connect('notify::color', self._on_color_property_changed)

    def _on_color_selected(self, color, button):
        if self._selected_button:
            self._selected_button._checkmark.remove_css_class("visible")
        self._selected_button = button
        button._checkmark.add_css_class("visible")
        self.set_property('color', color)
        self.emit('color-changed', color)

    def _on_more_colors_clicked(self, button):
        self._show_color_dialog(self._on_color_dialog_response)

    def _on_color_dialog_response(self, dialog, result):
        try:
            color = dialog.choose_rgba_finish(result)
            if self._selected_button:
                self._selected_button._checkmark.remove_css_class("visible")
                self._selected_button = None
            self.set_property('color', color)
            self.emit('color-changed', color)
        except Exception:
            pass

    def _on_color_property_changed(self, widget, pspec):
        self._update_selection()

    def _update_selection(self):
        current_color = self.get_property('color')
        for child in self:
            if hasattr(child, '_color'):
                if self._colors_match(child._color, current_color):
                    if self._selected_button:
                        self._selected_button._checkmark.remove_css_class("visible")
                    self._selected_button = child
                    child._checkmark.add_css_class("visible")
                    break

    def _colors_match(self, color1, color2):
        tolerance = 0.01
        return (abs(color1.red - color2.red) < tolerance and
                abs(color1.green - color2.green) < tolerance and
                abs(color1.blue - color2.blue) < tolerance and
                abs(color1.alpha - color2.alpha) < tolerance)

    def get_color(self):
        return self.get_property('color')

    def set_color(self, color):
        self.set_property('color', color)
        self.emit('color-changed', color)

    def set_color_list(self, colors):
        self._custom_colors = colors
        self._create_color_row()

class SimpleColorPicker(Gtk.Button, ColorPickerMixin):
    __gtype_name__ = 'GradiaSimpleColorPicker'

    color = GObject.Property(
        type=Gdk.RGBA,
        default=Gdk.RGBA(0.2, 0.4, 1.0, 1.0),
        flags=GObject.ParamFlags.READWRITE
    )

    icon_name = GObject.Property(
        type=str,
        default="",
        flags=GObject.ParamFlags.READWRITE
    )

    text = GObject.Property(
        type=str,
        default="",
        flags=GObject.ParamFlags.READWRITE
    )

    with_alpha = GObject.Property(
        type=bool,
        default=True,
        flags=GObject.ParamFlags.READWRITE
    )

    __gsignals__ = {
        'color-changed': (GObject.SignalFlags.RUN_FIRST, None, (Gdk.RGBA,))
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_has_frame(False)
        self.add_css_class('flat')
        self._custom_colors = None
        self._setup_ui()
        self._setup_bindings()
        self.connect('clicked', self._on_clicked)

    def _setup_ui(self):
        self.content_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        self.content_box.set_halign(Gtk.Align.CENTER)
        self.content_box.set_valign(Gtk.Align.CENTER)

        self.icon = Gtk.Image()
        self.icon.set_visible(False)

        self.label = Gtk.Label()
        self.label.set_visible(False)

        self.content_box.append(self.icon)
        self.content_box.append(self.label)

        self.set_child(self.content_box)

        self.popover = Gtk.Popover()
        self.popover.set_parent(self)
        self.popover.set_position(Gtk.PositionType.BOTTOM)
        self._setup_popover_content()

        self._update_content()
        self._update_color_style()

    def _setup_popover_content(self):
        popover_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)

        if self._custom_colors is not None:
            color_palette = self._custom_colors
        else:
            color_palette = self._get_base_colors(secondary=True)

        for color, name in color_palette:
            color_button = self._create_color_button(color, name)
            color_button.connect('clicked', lambda btn, c=color: self._on_color_selected(c))
            popover_box.append(color_button)

        more_colors_button = self._create_more_colors_button()
        more_colors_button.connect('clicked', self._on_more_colors_clicked)
        popover_box.append(more_colors_button)

        self.popover.set_child(popover_box)

    def _setup_bindings(self):
        self.connect('notify::color', self._on_color_property_changed)
        self.connect('notify::icon-name', self._on_icon_name_changed)
        self.connect('notify::text', self._on_text_changed)

    def _update_content(self):
        icon_name = self.get_property('icon-name')
        text = self.get_property('text')

        if icon_name:
            self.icon.set_from_icon_name(icon_name)
            self.icon.set_visible(True)
        else:
            self.icon.set_visible(False)

        if text:
            self.label.set_text(text)
            self.label.set_visible(True)
        else:
            self.label.set_visible(False)

        if not icon_name and not text:
            self.content_box.set_size_request(24, 24)
        else:
            self.content_box.set_size_request(-1, -1)

    def _update_color_style(self):
        color = self.get_property('color')

        ctx = self.get_style_context()
        if hasattr(self, "_color_css_provider"):
            ctx.remove_provider(self._color_css_provider)

        if color.alpha == 0:
            css = """
                button {
                    background-color: #b2b2b2;
                }
                """
            self.remove_css_class("transparent-simple-color-picker")
            self.add_css_class("transparent-simple-color-picker")
        else:
            css = f"""
                button {{
                    background-color: rgba({int(color.red * 255)}, {int(color.green * 255)}, {int(color.blue * 255)}, {color.alpha});
                }}
                """
            self.remove_css_class("transparent-simple-color-picker")

        provider = Gtk.CssProvider()
        provider.load_from_data(css.encode())
        ctx.add_provider(provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)
        self._color_css_provider = provider
        self.add_css_class("simple-color-picker")

        if is_light_color_rgba(color):
            self.icon.add_css_class("dark")
            self.label.add_css_class("dark")
        else:
            self.icon.remove_css_class("dark")
            self.label.remove_css_class("dark")

    def _on_clicked(self, button):
        self.popover.popup()

    def _on_color_selected(self, color):
        self.set_property('color', color)
        self.popover.popdown()
        self.emit('color-changed', color)

    def _on_more_colors_clicked(self, button):
        self.popover.popdown()
        self._show_color_dialog(self._on_color_dialog_response)

    def _on_color_dialog_response(self, dialog, result):
        try:
            color = dialog.choose_rgba_finish(result)
            self.set_property('color', color)
            self.emit('color-changed', color)
        except Exception:
            pass

    def _on_color_property_changed(self, widget, pspec):
        self._update_color_style()

    def _on_icon_name_changed(self, widget, pspec):
        self._update_content()

    def _on_text_changed(self, widget, pspec):
        self._update_content()

    def get_color(self):
        return self.get_property('color')

    def set_color(self, color, emit=True):
        self.set_property('color', color)
        if emit:
            self.emit('color-changed', color)

    def get_icon_name(self):
        return self.get_property('icon-name')

    def set_icon_name(self, icon_name):
        self.set_property('icon-name', icon_name)

    def get_text(self):
        return self.get_property('text')

    def set_text(self, text):
        self.set_property('text', text)

    def set_color_list(self, colors):
        self._custom_colors = colors
        self._setup_popover_content()

    def set_color_by_index(self, index, emit=True):
        if index == None:
            return
        palette = self._custom_colors if self._custom_colors is not None else self._get_base_colors(secondary=True)
        if 0 <= index < len(palette):
            color, _ = palette[index]
            self.set_color(color, emit=emit)

    def do_dispose(self):
        if self.popover:
            self.popover.unparent()
        super().do_dispose()




# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any
from gi.repository import Gtk, Gio, GLib, GObject

class AspectRatioButton(Gtk.Button):
    __gtype_name__ = "GradiaAspectRatioButton"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self.set_tooltip_text(_("Aspect Ratio"))
        self.add_css_class("flat")

        self.default_icon_name = "aspect-ratio-symbolic"
        self.icon = Gtk.Image.new_from_icon_name(self.default_icon_name)
        self.set_child(self.icon)

        self.popover = Gtk.Popover()
        self.popover.set_parent(self)
        self.popover.set_position(Gtk.PositionType.TOP)

        self.aspect_ratios = [
            (_("Free"), "aspect-ratio-free-symbolic", 0.0),
            (_("Square"), "aspect-ratio-square-symbolic", 1.0),
            ("4:3", "aspect-ratio-4to3-symbolic", 4.0 / 3.0),
            ("3:2", "aspect-ratio-3to2-symbolic", 3.0 / 2.0),
            ("5:4", "aspect-ratio-5to4-symbolic", 5.0 / 4.0),
            ("16:9", "aspect-ratio-16to9-symbolic", 16.0 / 9.0),
        ]

        self._setup_popover_content()
        self.connect("clicked", self._on_button_clicked)

    def _setup_popover_content(self) -> None:
        vbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=6)
        vbox.set_margin_top(6)
        vbox.set_margin_bottom(6)
        vbox.set_margin_start(6)
        vbox.set_margin_end(6)

        for label_text, icon_name, ratio_value in self.aspect_ratios:
            button = Gtk.Button()
            button.set_has_frame(False)

            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=12)
            hbox.set_halign(Gtk.Align.START)

            icon = Gtk.Image.new_from_icon_name(icon_name)
            icon.set_icon_size(Gtk.IconSize.NORMAL)
            hbox.append(icon)

            label = Gtk.Label(label=label_text)
            label.set_halign(Gtk.Align.START)
            hbox.append(label)

            button.set_child(hbox)
            button.connect("clicked", self._on_aspect_ratio_selected, ratio_value, icon_name)
            vbox.append(button)

        self.popover.set_child(vbox)

    def _on_button_clicked(self, button: Gtk.Button) -> None:
        self.popover.popup()

    def _on_aspect_ratio_selected(self, button: Gtk.Button, ratio_value: float, icon_name: str) -> None:
        self.icon.set_from_icon_name(icon_name)
        window = self.get_root()
        action = window.lookup_action('aspect-ratio-crop')
        variant = GLib.Variant('d', ratio_value)
        action.activate(variant)
        self.popover.popdown()

    def do_dispose(self) -> None:
        if self.popover:
            self.popover.unparent()
        super().do_dispose()




# Copyright (C) 2025 tfuxu, Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gdk

HexColor = str
RGBTuple = tuple[int, int, int]

def hex_to_rgba(hex_color: HexColor, alpha: float | None = None) -> Gdk.RGBA:
    """
    Converts hexadecimal color code to `Gdk.RGBA` object.

    NOTE: If you are looking for the raw representation of
    red, green and blue channels, use `hex_to_rgb()` method instead.
    """

    rgba = Gdk.RGBA()
    rgba.parse(hex_color)

    if alpha is not None:
        rgba.alpha = alpha

    return rgba

def rgba_to_hex(rgba: Gdk.RGBA) -> HexColor:
    """
    Converts `Gdk.RGBA` object to hexadecimal representation of red, green and
    blue channels.
    """

    r = int(rgba.red * 255)
    g = int(rgba.green * 255)
    b = int(rgba.blue * 255)
    a = int(rgba.alpha * 255)
    return f"#{r:02x}{g:02x}{b:02x}{a:02x}"

def hex_to_rgb(hex_color: HexColor) -> RGBTuple:
    """
    Converts hexadecimal color code to raw representation of red, green and
    blue channels.
    """

    hex_color = hex_color.lstrip('#')
    r, g, b = (int(hex_color[i:i+2], 16) for i in (0, 2, 4))
    return (r, g, b)

def has_visible_color(color):
    return any(c > 0 for c in color[:3]) or (len(color) > 3 and color[3] > 0)

def _calculate_luminance(r: float, g: float, b: float, a: float) -> float:
    if a == 0:
        return 255
    return 0.299 * r * 255 + 0.587 * g * 255 + 0.114 * b * 255

def is_light_color_hex(hex_color: str) -> bool:
    hex_color = hex_color.lstrip("#")
    if len(hex_color) == 6:
        r, g, b = [int(hex_color[i:i+2], 16)/255 for i in (0, 2, 4)]
        a = 1.0
    elif len(hex_color) == 8:
        r, g, b, a = [int(hex_color[i:i+2], 16)/255 for i in (0, 2, 4, 6)]
    else:
        raise ValueError("Invalid hex color format")
    return _calculate_luminance(r, g, b, a) > 130

def is_light_color_rgba(rgba: Gdk.RGBA) -> bool:
    return _calculate_luminance(rgba.red, rgba.green, rgba.blue, rgba.alpha) > 130

def parse_rgb_string(s: str) -> tuple[int, int, int]:
    s = s.strip().lower()
    if s.startswith("rgb(") and s.endswith(")"):
        parts = s[4:-1].split(",")
        return tuple(int(p.strip()) for p in parts)
    raise ValueError(f"Invalid rgb string: {s}")




# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later
from typing import Optional

from PIL import Image, ImageChops, ImageDraw, ImageFilter
from gi.repository import GdkPixbuf

from gradia.graphics.background import Background
from gradia.graphics.loaded_image import LoadedImage, BalancedPadding


class ImageProcessor:
    def __init__(
        self,
        image: Optional[LoadedImage] = None,
        background: Optional[Background] = None,
        padding: int = 0,
        aspect_ratio: Optional[str | float] = None,
        corner_radius: int = 0,
        shadow_strength: float = 0,
        auto_balance: bool = False,
        rotation: int = 0
    ) -> None:
        self.background: Optional[Background] = background
        self.padding: int = padding
        self.shadow_strength: float = shadow_strength
        self.aspect_ratio: Optional[str | float] = aspect_ratio
        self.corner_radius: int = corner_radius
        self.auto_balance: bool = auto_balance
        self.rotation: int = rotation
        self._loaded_image: Optional[LoadedImage] = None

        if image:
            self.set_image(image)

    def set_image(self, image: LoadedImage) -> None:
        if not image.is_loaded:
            raise ValueError(f"Failed to load image: {image.load_error}")
        self._loaded_image = image

    def process_to_pillow(self) -> Image:
        if not self._loaded_image or not self._loaded_image.preview_image:
            raise ValueError("No image loaded to process")
        source_img = self._loaded_image.preview_image.copy()

        if self.rotation != 0:
            source_img = self._apply_rotation(source_img)
        width, height = source_img.size

        if self.auto_balance and self._loaded_image.balanced_padding:
            source_img = self._apply_auto_balance(source_img, self._loaded_image.balanced_padding)
            width, height = source_img.size

        if self.padding < 0:
            source_img = self._crop_image(source_img)
            width, height = source_img.size

        if self.corner_radius > 0:
            source_img = self._apply_rounded_corners(source_img)

        padded_width, padded_height = self._calculate_final_dimensions(width, height)
        final_img = self._create_background(padded_width, padded_height)
        paste_position = self._get_paste_position(width, height, padded_width, padded_height)

        shadow_img, shadow_offset = self._create_shadow(source_img, offset=(10, 10), shadow_strength=self.shadow_strength)
        shadow_position = (paste_position[0] - shadow_offset[0], paste_position[1] - shadow_offset[1])

        final_img = self._alpha_composite_at_position(final_img, shadow_img, shadow_position)
        final_img = self._alpha_composite_at_position(final_img, source_img, paste_position)

        return final_img

    def process(self) -> tuple[GdkPixbuf.Pixbuf, int, int]:
        final_img = self.process_to_pillow()
        final_pixbuf = self._pil_to_pixbuf(final_img)
        full_width, full_height = self.get_full_resolution_dimensions(final_pixbuf)
        return final_pixbuf, full_width, full_height

    def process_full_resolution(self) -> GdkPixbuf.Pixbuf:
        if not self._loaded_image or not self._loaded_image.full_res_image:
            raise ValueError("No full resolution image loaded to process")

        source_img = self._loaded_image.full_res_image.copy()

        if self.rotation != 0:
            source_img = self._apply_rotation(source_img)

        width, height = source_img.size

        if self.auto_balance and self._loaded_image.balanced_padding:
            source_img = self._apply_auto_balance_full_res(source_img, self._loaded_image.balanced_padding)
            width, height = source_img.size

        if self.padding < 0:
            source_img = self._crop_image(source_img)
            width, height = source_img.size

        if self.corner_radius > 0:
            source_img = self._apply_rounded_corners(source_img)

        padded_width, padded_height = self._calculate_final_dimensions(width, height)
        final_img = self._create_background_full_res(padded_width, padded_height)
        paste_position = self._get_paste_position(width, height, padded_width, padded_height)

        shadow_img, shadow_offset = self._create_shadow_full_res(source_img, offset=(10, 10), shadow_strength=self.shadow_strength)
        shadow_position = (paste_position[0] - shadow_offset[0], paste_position[1] - shadow_offset[1])
        final_img = self._alpha_composite_at_position(final_img, shadow_img, shadow_position)

        final_img = self._alpha_composite_at_position(final_img, source_img, paste_position)

        return self._pil_to_pixbuf(final_img)

    def _apply_rotation(self, image: Image.Image) -> Image.Image:
        if self.rotation == 0:
            return image
        elif self.rotation == 90:
            return image.transpose(Image.ROTATE_270)
        elif self.rotation == 180:
            return image.transpose(Image.ROTATE_180)
        elif self.rotation == 270:
            return image.transpose(Image.ROTATE_90)
        else:
            return image

    def _apply_auto_balance(self, image: Image.Image, balanced_padding: BalancedPadding) -> Image.Image:
        width, height = image.size

        new_width = width + balanced_padding.left + balanced_padding.right
        new_height = height + balanced_padding.top + balanced_padding.bottom

        balanced_image = Image.new("RGBA", (new_width, new_height), balanced_padding.color)

        paste_x = balanced_padding.left
        paste_y = balanced_padding.top
        balanced_image.paste(image, (paste_x, paste_y), image)

        return balanced_image

    def _apply_auto_balance_full_res(self, image: Image.Image, balanced_padding: BalancedPadding) -> Image.Image:
        if not self._loaded_image or not self._loaded_image.preview_image or not self._loaded_image.full_res_image:
            return image

        scale_x = self._loaded_image.full_res_image.width / self._loaded_image.preview_image.width
        scale_y = self._loaded_image.full_res_image.height / self._loaded_image.preview_image.height
        scale = max(scale_x, scale_y)

        width, height = image.size
        top = int(balanced_padding.top * scale)
        bottom = int(balanced_padding.bottom * scale)
        left = int(balanced_padding.left * scale)
        right = int(balanced_padding.right * scale)

        new_width = width + left + right
        new_height = height + top + bottom

        balanced_image = Image.new("RGBA", (new_width, new_height), balanced_padding.color)

        paste_x = left
        paste_y = top
        balanced_image.paste(image, (paste_x, paste_y), image)

        return balanced_image

    def _create_background_full_res(self, width: int, height: int) -> Image.Image:
        if self.background:
            return self.background.prepare_image(width, height)
        return Image.new("RGBA", (width, height), (0, 0, 0, 0))

    def _create_shadow_full_res(
        self,
        image: Image.Image,
        offset: tuple[int, int] = (10, 10),
        shadow_strength: float = 1.0
    ) -> tuple[Image.Image, tuple[int, int]]:
        if not self._loaded_image or not self._loaded_image.preview_image or not self._loaded_image.full_res_image:
            scale = 1.0
        else:
            scale_x = self._loaded_image.full_res_image.width / self._loaded_image.preview_image.width
            scale_y = self._loaded_image.full_res_image.height / self._loaded_image.preview_image.height
            scale = max(scale_x, scale_y)

        shadow_strength = max(0.0, min(shadow_strength, 10)) / 5
        blur_radius = int(10 * shadow_strength * scale)
        shadow_alpha = int(150 * shadow_strength)
        shadow_color = (0, 0, 0, shadow_alpha)
        scaled_offset = (int(offset[0] * scale), int(offset[1] * scale))

        alpha = image.split()[3]
        shadow = Image.new("RGBA", image.size, shadow_color)
        shadow.putalpha(alpha)

        extra_margin = blur_radius * 5
        expanded_width = image.width + abs(scaled_offset[0]) + extra_margin
        expanded_height = image.height + abs(scaled_offset[1]) + extra_margin
        shadow_canvas = Image.new("RGBA", (expanded_width, expanded_height), (0, 0, 0, 0))

        shadow_x = extra_margin // 2 + max(scaled_offset[0], 0)
        shadow_y = extra_margin // 2 + max(scaled_offset[1], 0)
        shadow_canvas.paste(shadow, (shadow_x, shadow_y), shadow)

        shadow_canvas = shadow_canvas.filter(ImageFilter.GaussianBlur(blur_radius))

        return shadow_canvas, (shadow_x, shadow_y)

    def _get_percentage(self, value: float) -> float:
        return value / 100.0

    def _alpha_composite_at_position(
        self,
        background: Image.Image,
        foreground: Image.Image,
        position: tuple[int, int]
    ) -> Image.Image:
        if background.mode != 'RGBA':
            background = background.convert('RGBA')

        if foreground.mode != 'RGBA':
            foreground = foreground.convert('RGBA')

        temp_canvas = Image.new('RGBA', background.size, (0, 0, 0, 0))
        temp_canvas.paste(foreground, position, foreground)
        result = Image.alpha_composite(background, temp_canvas)

        return result

    def _crop_image(self, image: Image.Image) -> Image.Image:
        width, height = image.size
        smaller_dimension = min(width, height)
        padding_percentage = self._get_percentage(abs(self.padding))
        padding_pixels = int(padding_percentage * smaller_dimension)

        crop_width = max(1, width - 2 * padding_pixels)
        crop_height = max(1, height - 2 * padding_pixels)
        offset_x = (width - crop_width) // 2
        offset_y = (height - crop_height) // 2

        return image.crop(
            (offset_x, offset_y, offset_x + crop_width, offset_y + crop_height)
        )

    def get_full_resolution_dimensions(self, processed_pixbuf) -> tuple[int, int]:
        if not self._loaded_image or not self._loaded_image.full_res_image or not self._loaded_image.preview_image:
            raise ValueError("No images loaded")

        scale_x = self._loaded_image.full_res_image.width / self._loaded_image.preview_image.width
        scale_y = self._loaded_image.full_res_image.height / self._loaded_image.preview_image.height
        scale = max(scale_x, scale_y)

        downscaled_width = processed_pixbuf.get_width()
        downscaled_height = processed_pixbuf.get_height()

        full_width = int(downscaled_width * scale)
        full_height = int(downscaled_height * scale)

        return full_width, full_height

    def _calculate_final_dimensions(self, width: int, height: int) -> tuple[int, int]:
       width = max(1, width)
       height = max(1, height)

       if self.padding >= 0:
           smaller_dimension = min(width, height)
           padding_percentage = self._get_percentage(self.padding)
           padding_pixels = int(padding_percentage * smaller_dimension)
           width += padding_pixels * 2
           height += padding_pixels * 2

       if self.aspect_ratio:
           width, height = self._adjust_for_aspect_ratio(width, height)

       return width, height

    def _adjust_for_aspect_ratio(self, width: int, height: int) -> tuple[int, int]:
        try:
            ratio = self._parse_aspect_ratio()
            current = width / height

            if current < ratio:
                width = int(height * ratio)
            elif current > ratio:
                height = int(width / ratio)

            return width, height
        except Exception:
            return width, height

    def _parse_aspect_ratio(self) -> float:
        if isinstance(self.aspect_ratio, str) and ":" in self.aspect_ratio:
            w, h = map(float, self.aspect_ratio.split(":"))
            return w / h

        if self.aspect_ratio:
            return float(self.aspect_ratio)

        raise ValueError("aspect_ratio is None and cannot be converted to float")

    def _apply_rounded_corners(self, image: Image.Image) -> Image.Image:
        width, height = image.size
        smaller_dimension = min(width, height)
        radius_percentage = self._get_percentage(self.corner_radius)
        radius_pixels = int(radius_percentage * smaller_dimension)
        oversample = 4

        large_mask = Image.new("L", (width * oversample, height * oversample), 0)

        draw = ImageDraw.Draw(large_mask)
        draw.rounded_rectangle(
            (0, 0, width * oversample, height * oversample),
            radius=radius_pixels * oversample,
            fill=255
        )

        mask = large_mask.resize((width, height), Image.LANCZOS)
        r, g, b, alpha = image.split()
        new_alpha = ImageChops.multiply(alpha, mask)

        return Image.merge("RGBA", (r, g, b, new_alpha))

    def _create_background(self, width: int, height: int) -> Image.Image:
        if self.background:
            return self.background.prepare_image(width, height)
        return Image.new("RGBA", (width, height), (0, 0, 0, 0))

    def _create_shadow(
        self,
        image: Image.Image,
        offset: tuple[int, int] = (10, 10),
        shadow_strength: float = 1.0
    ) -> tuple[Image.Image, tuple[int, int]]:
        shadow_strength = max(0.0, min(shadow_strength, 10)) / 5
        blur_radius = int(10 * shadow_strength)
        shadow_alpha = int(150 * shadow_strength)
        shadow_color = (0, 0, 0, shadow_alpha)

        alpha = image.split()[3]
        shadow = Image.new("RGBA", image.size, shadow_color)
        shadow.putalpha(alpha)

        extra_margin = blur_radius * 5
        expanded_width = image.width + abs(offset[0]) + extra_margin
        expanded_height = image.height + abs(offset[1]) + extra_margin
        shadow_canvas = Image.new("RGBA", (expanded_width, expanded_height), (0, 0, 0, 0))

        shadow_x = extra_margin // 2 + max(offset[0], 0)
        shadow_y = extra_margin // 2 + max(offset[1], 0)
        shadow_canvas.paste(shadow, (shadow_x, shadow_y), shadow)

        shadow_canvas = shadow_canvas.filter(ImageFilter.GaussianBlur(blur_radius))

        return shadow_canvas, (shadow_x, shadow_y)

    def _get_paste_position(
        self,
        image_width: int,
        image_height: int,
        background_width: int,
        background_height: int
    ) -> tuple[int, int]:
        if self.padding >= 0:
            x = (background_width - image_width) // 2
            y = (background_height - image_height) // 2
            return x, y
        return 0, 0

    def _pil_to_pixbuf(self, image: Image.Image) -> GdkPixbuf.Pixbuf:
        if image.mode != 'RGBA':
            image = image.convert('RGBA')

        width, height = image.size
        pixels = image.tobytes()

        pixbuf = GdkPixbuf.Pixbuf.new_from_data(
            data=pixels,
            colorspace=GdkPixbuf.Colorspace.RGB,
            has_alpha=True,
            bits_per_sample=8,
            width=width,
            height=height,
            rowstride=width * 4,
            destroy_fn=None
        )

        return pixbuf



# Copyright (C) 2025 Alexander Vanhee
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import mimetypes
import shutil
import threading
from urllib.parse import urlparse, unquote
import urllib.request
from gi.repository import Gtk, Gio, Gdk, GLib, Xdp
from gradia.clipboard import save_texture_to_file
from gradia.ui.image_creation.source_image_generator import SourceImageGeneratorWindow
from gradia.utils.timestamp_filename import TimestampedFilenameGenerator
from gradia.backend.logger import Logger
from gradia.graphics.loaded_image import LoadedImage, ImageOrigin
from typing import Optional, Callable
ImportFormat = tuple[str, str]

logger = Logger()

class BaseImageLoader:
    SUPPORTED_INPUT_FORMATS: list[ImportFormat] = [
        (".png", "image/png"),
        (".jpg", "image/jpg"),
        (".jpeg", "image/jpeg"),
        (".webp", "image/webp"),
        (".avif", "image/avif"),
    ]

    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str) -> None:
        self.window: Gtk.ApplicationWindow = window
        self.temp_dir: str = temp_dir

    def _is_supported_format(self, file_path: str) -> bool:
        lower_path = file_path.lower()
        supported_extensions = [ext for ext, _ in self.SUPPORTED_INPUT_FORMATS]
        return any(lower_path.endswith(ext) for ext in supported_extensions)

    def _set_image_and_update_ui(self, file_path: str, origin: ImageOrigin, screenshot_path: str = None, copy_after_processing: bool = False) -> None:
        self.window.show_loading_state()

        def load_image_thread():
            try:
                loaded_image = LoadedImage(file_path, origin, screenshot_path)
                GLib.idle_add(self._on_image_loaded, loaded_image, copy_after_processing)
            except Exception as e:
                logger.error(f"Error loading image in thread: {e}")
                GLib.idle_add(self._on_image_load_error, str(e))

        thread = threading.Thread(target=load_image_thread, daemon=True)
        thread.start()

    def _on_image_loaded(self, loaded_image: LoadedImage, copy_after_processing: bool) -> bool:
        self.window.set_image(loaded_image, copy_after_processing=copy_after_processing)
        return False

    def _on_image_load_error(self, error_message: str) -> bool:
        self.window._show_notification(f"Failed to load image: {error_message}")
        self.window.hide_loading_state()
        return False

    def _handle_uri(self, uri: str, origin: ImageOrigin) -> bool:
        logger.info(f"Processing URI: {uri}")

        if uri.startswith("file://"):
            file_path = unquote(urlparse(uri).path)

            if not os.path.isfile(file_path):
                logger.error("File does not exist:", file_path)
                return False

            if not self._is_supported_format(file_path):
                self.window._show_notification(_("Not a supported image format"))
                return False

            self._set_image_and_update_ui(file_path, origin)
            return True

        elif uri.startswith(("http://", "https://")):
            return self._handle_image_url(uri, origin)

        else:
            logger.info("Unsupported URI scheme:", uri)
            self.window._show_notification(_("Not a supported image format"))
            return False

    def _handle_image_url(self, url: str, origin: ImageOrigin) -> bool:
        try:
            path = urlparse(url).path
            mime_type, _unused = mimetypes.guess_type(path)
            logger.info(f"mime type from guess_type: {mime_type}")

            if not (mime_type and mime_type.startswith("image/")):
                lower_path = path.lower()
                supported_extensions = [ext for ext, _ in self.SUPPORTED_INPUT_FORMATS]
                if not any(lower_path.endswith(ext) for ext in supported_extensions):
                    self.window._show_notification(_("Not a supported image format"))
                    return False
                else:
                    logger.info("Fallback: file extension matches supported image format.")

            filename = os.path.basename(path) or "downloaded_image"
            temp_path = os.path.join(self.temp_dir, filename)

            urllib.request.urlretrieve(url, temp_path)

            if not self._is_supported_format(temp_path):
                self.window._show_notification(_("URL is not a supported image format"))
                os.remove(temp_path)
                return False

            self._set_image_and_update_ui(temp_path, origin)
            return True

        except Exception as e:
            logger.error("Error downloading image:", e)
            self.window._show_notification(_("Failed to load image from URL."))
            return False

class FileDialogImageLoader(BaseImageLoader):
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str) -> None:
        super().__init__(window, temp_dir)

    def open_file_dialog(self) -> None:
        file_dialog = Gtk.FileDialog()
        file_dialog.set_title(_("Open Image"))

        image_filter = Gtk.FileFilter()
        image_filter.set_name(_("Image Files"))
        for _ext, mime_type in self.SUPPORTED_INPUT_FORMATS:
            image_filter.add_mime_type(mime_type)

        filters = Gio.ListStore.new(Gtk.FileFilter)
        filters.append(image_filter)
        file_dialog.set_filters(filters)

        file_dialog.open(self.window, None, self._on_file_selected)

    def _on_file_selected(self, dialog: Gtk.FileDialog, result: Gio.AsyncResult) -> None:
        try:
            file = dialog.open_finish(result)
            if not file:
                return

            file_path = file.get_path()
            if not file_path or not os.path.isfile(file_path):
                logger.info(f"Invalid file path: {file_path}")
                return

            if not self._is_supported_format(file_path):
                logger.info(f"Unsupported file format: {file_path}")
                return

            self._set_image_and_update_ui(file_path, ImageOrigin.FileDialog)

        except Exception as e:
            logger.error(f"Error opening file: {e}")


class DragDropImageLoader(BaseImageLoader):
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str) -> None:
        super().__init__(window, temp_dir)

    def handle_file_drop(
        self,
        drop_target: Optional[object],
        value: object,
        x: int,
        y: int
    ) -> bool:

        if isinstance(value, Gio.File):
            uri = value.get_uri()
            return self._handle_uri(uri, ImageOrigin.DragDrop)

        return False


class ClipboardImageLoader(BaseImageLoader):
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str) -> None:
        super().__init__(window, temp_dir)

    def load_from_clipboard(self) -> None:
        display = Gdk.Display.get_default()
        clipboard = display.get_clipboard()

        clipboard.read_value_async(
            Gdk.FileList,
            0,
            None,
            self._on_uris_get
        )

    def _on_uris_get(self, clipboard, result, user_data=None) -> None:
        try:
            file_list = clipboard.read_value_finish(result)
            if file_list:
                files = file_list.get_files()
                if files:
                    file_obj = files[0]
                    uri = file_obj.get_uri()
                    self._handle_uri(uri, ImageOrigin.Clipboard)
        except GLib.GError as e:
            print(f"Error reading URIs: {e}")


class ScreenshotImageLoader(BaseImageLoader):
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str, app: Gtk.Application) -> None:
        super().__init__(window, temp_dir)
        self.portal = Xdp.Portal()
        self._error_callback: Optional[Callable[[str], None]] = None
        self._success_callback: Optional[Callable[[], None]] = None
        self._screenshot_uris: list[str] = []
        self.window = window

    def _update_delete_action_state(self) -> None:
        action = self.window.lookup_action("delete-screenshots")
        if action:
            action.set_enabled(bool(self._screenshot_uris))

    def take_screenshot(
        self,
        flags: Xdp.ScreenshotFlags = Xdp.ScreenshotFlags.INTERACTIVE,
        on_error_or_cancel: Optional[Callable[[str], None]] = None,
        on_success: Optional[Callable[[], None]] = None
    ) -> None:
        try:
            self._error_callback = on_error_or_cancel
            self._success_callback = on_success
            self.window.hide()
            GLib.timeout_add(150, self._do_take_screenshot, flags)
        except Exception as e:
            logger.error(f"Failed to initiate screenshot: {e}")
            self.window._show_notification(_("Failed to take screenshot"))
            if on_error_or_cancel:
                on_error_or_cancel(str(e))

    def _do_take_screenshot(self, flags: Xdp.ScreenshotFlags) -> bool:
        try:
            self.portal.take_screenshot(
                None,
                flags,
                None,
                self._on_screenshot_taken,
                None
            )
        except Exception as e:
            logger.info(f"Failed during screenshot: {e}")
            self.window._show_notification(_("Failed to take screenshot"))
            self.window.show()
            if self._error_callback:
                self._error_callback(str(e))
            self._error_callback = None
        return False

    def _on_screenshot_taken(self, portal_object, result, user_data) -> None:
        try:
            uri = self.portal.take_screenshot_finish(result)
            self._screenshot_uris.append(uri)
            self._handle_screenshot_uri(uri)
            self._update_delete_action_state()
        except GLib.Error as e:
            logger.error(f"Screenshot error: {e}")
            self.window._show_notification(_("Screenshot cancelled"))
            if self._error_callback:
                self._error_callback(str(e))
        finally:
            self.window.show()
            self._error_callback = None

    def _handle_screenshot_uri(self, uri: str) -> None:
        try:
            file = Gio.File.new_for_uri(uri)
            original_path = file.get_path()
            success, contents, _unused = file.load_contents(None)
            if not success or not contents:
                raise Exception("Failed to load screenshot data")

            filename = TimestampedFilenameGenerator().generate(_("Edited Screenshot From %Y-%m-%d %H-%M-%S")) + ".png"
            temp_path = os.path.join(self.temp_dir, filename)

            with open(temp_path, 'wb') as f:
                f.write(contents)

            self._set_image_and_update_ui(temp_path, ImageOrigin.Screenshot, screenshot_path=original_path, copy_after_processing=True)
            self.window._show_notification(_("Screenshot captured!"))

            if self._success_callback:
                self._success_callback()

        except Exception as e:
            logger.error(f"Error processing screenshot: {e}")
            self.window._show_notification(_("Failed to process screenshot"))

    def load_path_as_screenshot(self, file_path: str) -> None:
        try:
            file = Gio.File.new_for_path(file_path)
            uri = file.get_uri()
            self._screenshot_uris.append(uri)
            self._update_delete_action_state()

            filename = TimestampedFilenameGenerator().generate(_("Edited Screenshot From %Y-%m-%d %H-%M-%S")) + ".png"
            new_path = os.path.join(self.temp_dir, filename)

            shutil.copy(file_path, new_path)

            self._set_image_and_update_ui(file_path, ImageOrigin.FakeScreenshot, screenshot_path=file_path, copy_after_processing=True)

            self.window._show_notification(_("Screenshot captured!"))

        except Exception as e:
            logger.error(f"Error loading screenshot from path: {e}")
            self.window._show_notification(_("Failed to load screenshot"))

    def get_screenshot_uris(self) -> list[str]:
        return self._screenshot_uris.copy()

    def delete_screenshots(self) -> None:
        for uri in self._screenshot_uris:
            try:
                file = Gio.File.new_for_uri(uri)
                file.trash(None)
            except Exception as e:
                logger.error(f"Failed to trash screenshot {uri}: {e}")

        self._screenshot_uris.clear()
        self._update_delete_action_state()


class CommandlineLoader(BaseImageLoader):
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str) -> None:
        super().__init__(window, temp_dir)

    def load_from_file(self, file_path: str) -> None:
        try:
            if not file_path:
                logger.info("No file path provided")
                return

            if not os.path.isfile(file_path):
                logger.info(f"File does not exist: {file_path}")
                return

            if not self._is_supported_format(file_path):
                logger.info(f"Unsupported file format: {file_path}")
                return

            self._set_image_and_update_ui(file_path, ImageOrigin.CommandLine)

        except Exception as e:
            logger.error(f"Error loading file from command line: {e}")

class SourceImageLoader(BaseImageLoader):
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str) -> None:
        super().__init__(window, temp_dir)
        self._generator_window: Optional[SourceImageGeneratorWindow] = None

    def open_generator(self) -> None:
        if self._generator_window and self._generator_window.get_visible():
            self._generator_window.present()
            return

        self._generator_window = SourceImageGeneratorWindow(parent_window=self.window, temp_dir=self.temp_dir, export_callback=self.load_generated_image)
        self._generator_window.set_transient_for(self.window)
        self._generator_window.connect("destroy", self._on_generator_window_destroyed)
        self._generator_window.show()

    def _on_generator_window_destroyed(self, window: Gtk.Window) -> None:
        self._generator_window = None

    def load_generated_image(self, image_path: str) -> None:
        if not image_path or not os.path.isfile(image_path):
            logger.warning(f"Invalid generated image path: {image_path}")
            return

        self._set_image_and_update_ui(image_path, ImageOrigin.SourceImage)
        self.window._show_notification(_("Source snippet Generated!"))

class ImportManager:
    def __init__(self, window: Gtk.ApplicationWindow, temp_dir: str, app: Gtk.Application) -> None:
        self.window: Gtk.ApplicationWindow = window
        self.temp_dir: str = temp_dir

        self.file_loader: FileDialogImageLoader = FileDialogImageLoader(window, temp_dir)
        self.drag_drop_loader: DragDropImageLoader = DragDropImageLoader(window, temp_dir)
        self.clipboard_loader: ClipboardImageLoader = ClipboardImageLoader(window, temp_dir)
        self.screenshot_loader: ScreenshotImageLoader = ScreenshotImageLoader(window, temp_dir, app)
        self.commandline_loader: CommandlineLoader = CommandlineLoader(window, temp_dir)
        self.source_image_loader: SourceImageLoader = SourceImageLoader(window, temp_dir)

    def open_file_dialog(self) -> None:
        self.file_loader.open_file_dialog()

    def _on_drop_action(self, action: Optional[object], param: object) -> None:
        if isinstance(param, GLib.Variant):
            uri = param.get_string()
            file = Gio.File.new_for_uri(uri)
            self.drag_drop_loader.handle_file_drop(None, file, 0, 0)
        else:
            logger.info("ImportManager._on_drop_action: Invalid drop parameter")

    def load_from_clipboard(self) -> None:
        self.clipboard_loader.load_from_clipboard()

    def take_screenshot(
        self,
        flags: Xdp.ScreenshotFlags = Xdp.ScreenshotFlags.INTERACTIVE,
        on_error_or_cancel: Optional[Callable[[str], None]] = None,
        on_success: Optional[Callable[[], None]] = None
    ) -> None:
        self.screenshot_loader.take_screenshot(flags, on_error_or_cancel, on_success)

    def load_as_screenshot(self, file_path: str):
        self.screenshot_loader.load_path_as_screenshot(file_path)

    def get_screenshot_uris(self) -> list[str]:
        return self.screenshot_loader.get_screenshot_uris()

    def delete_screenshots(self) -> None:
        return self.screenshot_loader.delete_screenshots()

    def load_from_file(self, file_path: str) -> None:
        self.commandline_loader.load_from_file(file_path)

    def generate_from_source_code(self) -> None:
        self.source_image_loader.open_generator()



# Copyright (C) 2025 Alexander Vanhee, tfuxu
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# SPDX-License-Identifier: GPL-3.0-or-later

from collections.abc import Callable
from typing import Optional
import json
from PIL import Image
from gi.repository import Adw, Gtk, Gio, Gdk

from gradia.graphics.background import Background
from gradia.utils.colors import hex_to_rgb, hex_to_rgba, rgba_to_hex, is_light_color_hex
from gradia.constants import rootdir  # pyright: ignore


class SolidBackground(Background):
    def __init__(self, color: str = "#4A90E2", alpha: float = 1.0) -> None:
        self.color = color
        self.alpha = alpha

    @classmethod
    def from_json(cls, json_str: str) -> 'SolidBackground':
        data = json.loads(json_str)
        return cls(
            color=data.get("color", "#4A90E2"),
            alpha=data.get("alpha", 1.0)
        )

    def to_json(self) -> str:
        return json.dumps({
            "type": "solid",
            "color": self.color,
            "alpha": self.alpha
        })

    def get_name(self) -> str:
        return f"solid-{self.color}-{self.alpha}"

    def prepare_image(self, width: int, height: int) -> Image.Image:
        rgb = hex_to_rgb(self.color)
        alpha_value = int(self.alpha * 255)
        return Image.new('RGBA', (width, height), (*rgb, alpha_value))


class ColorPresetButton(Gtk.Button):
    def __init__(self, color: str, alpha: float = 1.0, tooltip_text: str = "", **kwargs) -> None:
        super().__init__(
            valign=Gtk.Align.CENTER,
            width_request=40,
            height_request=40,
            margin_top=6,
            margin_bottom=6,
            margin_start=6,
            margin_end=6,
            tooltip_text=tooltip_text,
            **kwargs
        )
        self.set_focusable(True)
        self.set_can_focus(True)
        self.color = color
        self.alpha = alpha
        self.is_selected = False
        self._apply_style()
        self._setup_checkmark()

    def _apply_style(self) -> None:
        context = self.get_style_context()
        context.add_class("color-button")
        if self.alpha == 0.0:
            context.add_class("transparent-color-button")
        else:
            rgba = hex_to_rgba(self.color, self.alpha)
            provider = Gtk.CssProvider()
            provider.load_from_string(f"""
                button {{
                    background-color: {rgba.to_string()};
                }}
            """)
            context.add_provider(provider, Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def _setup_checkmark(self) -> None:
        self.checkmark = Gtk.Image.new_from_icon_name("object-select-symbolic")
        self.checkmark.set_pixel_size(16)
        self.checkmark.add_css_class("checkmark-icon")

        if is_light_color_hex(self.color) or self.alpha == 0:
            self.checkmark.add_css_class("dark")
        else:
             self.checkmark.remove_css_class("dark")

        overlay = Gtk.Overlay(width_request=16, height_request=16)
        overlay.set_child(Gtk.Box())
        overlay.add_overlay(self.checkmark)
        overlay.set_halign(Gtk.Align.CENTER)
        overlay.set_valign(Gtk.Align.CENTER)

        self.set_child(overlay)

    def set_selected(self, selected: bool) -> None:
        self.is_selected = selected
        if selected:
             self.checkmark.add_css_class("visible")
        else:
            self.checkmark.remove_css_class("visible")


class ColorPickerButton(Gtk.Button):
    def __init__(self, callback: Optional[Callable[[str, float], None]] = None, **kwargs) -> None:
        super().__init__(
            valign=Gtk.Align.CENTER,
            width_request=30,
            height_request=30,
            margin_top=6,
            margin_bottom=6,
            margin_start=6,
            margin_end=6,
            tooltip_text=_("More Colors"),
            **kwargs
        )
        self.set_focusable(True)
        self.set_can_focus(True)
        self.callback = callback
        self.is_selected = False
        self._setup_icon()
        self._apply_style()

    def _setup_icon(self) -> None:
        self.checkmark = Gtk.Image.new_from_icon_name("object-select-symbolic")
        self.checkmark.set_pixel_size(16)
        self.checkmark.add_css_class("checkmark-icon")


        overlay = Gtk.Overlay(width_request=40, height_request=40)
        overlay.add_overlay(self.checkmark)
        overlay.set_halign(Gtk.Align.CENTER)
        overlay.set_valign(Gtk.Align.CENTER)

        self.set_child(overlay)

    def _apply_style(self) -> None:
        context = self.get_style_context()
        context.add_class("color-button")
        context.add_class("color-picker-button")

    def set_selected(self, selected: bool) -> None:
        self.is_selected = selected
        if selected:
            self.checkmark.add_css_class("visible")
        else:
            self.checkmark.remove_css_class("visible")

    def open_color_picker(self) -> None:
        dialog = Gtk.ColorDialog()
        dialog.choose_rgba(
            parent=self.get_root(),
            initial_color=Gdk.RGBA(red=0.5, green=0.5, blue=0.5, alpha=1.0),
            cancellable=None,
            callback=self._on_color_chosen
        )

    def _on_color_chosen(self, dialog: Gtk.ColorDialog, result: Gio.AsyncResult) -> None:
        try:
            rgba = dialog.choose_rgba_finish(result)
            color_hex = rgba_to_hex(rgba)
            alpha = rgba.alpha

            if self.callback:
                self.callback(color_hex, alpha)
        except Exception as e:
            print(f"Color picker cancelled or error: {e}")


@Gtk.Template(resource_path=f"{rootdir}/ui/selectors/solid_selector.ui")
class SolidSelector(Adw.PreferencesGroup):
    __gtype_name__ = "GradiaSolidSelector"
    COMMON_COLORS = [
        ("#fff66151", _("Red")),
        ("#ffe66100", _("Orange")),
        ("#fff6d32d", _("Yellow")),
        ("#ff33d17a", _("Green")),
        ("#ff3584e4", _("Blue")),
        ("#ffc061cb", _("Purple")),
        ("#ffffd1dc", _("Pastel Pink")),
        ("#fffff4d1", _("Pastel Yellow")),
        ("#ffd1ffd1", _("Pastel Green")),
        ("#ffd1f0ff", _("Pastel Cyan")),
        ("#ffffffff", _("White")),
        ("#ff000000", _("Black")),
        ("#ff77767b", _("Gray")),
        ("#00000000", _("Transparent")),
    ]

    color_presets_grid: Gtk.Grid = Gtk.Template.Child()

    def __init__(
        self,
        solid: SolidBackground,
        callback: Optional[Callable[[SolidBackground], None]] = None,
        **kwargs
    ) -> None:
        super().__init__(**kwargs)
        self.solid = solid
        self.callback = callback
        self.preset_buttons = []
        self.color_picker_button = None
        self.custom_color = None
        self.custom_alpha = 1.0
        self._setup_color_presets_row()
        self._update_selected_preset()

    def _setup_color_presets_row(self) -> None:
        columns = 5
        for index, (color, color_name) in enumerate(self.COMMON_COLORS):
            hex_color = color.lstrip('#')
            if len(hex_color) == 8:
                alpha_from_hex = int(hex_color[:2], 16) / 255.0
                rgb_hex = hex_color[2:]
            else:
                alpha_from_hex = 1.0
                rgb_hex = hex_color

            full_color = f"#{rgb_hex}"
            button = ColorPresetButton(full_color, alpha_from_hex, color_name)
            button.connect("clicked", self._on_common_color_clicked, full_color, alpha_from_hex)

            row_pos = index // columns
            col_pos = index % columns
            self.color_presets_grid.attach(button, col_pos, row_pos, 1, 1)
            self.preset_buttons.append(button)

        self.color_picker_button = ColorPickerButton(self._on_custom_color_picked)
        self.color_picker_button.connect("clicked", self._on_color_picker_clicked)

        row_pos = len(self.COMMON_COLORS) // columns
        col_pos = len(self.COMMON_COLORS) % columns
        self.color_presets_grid.attach(self.color_picker_button, col_pos, row_pos, 1, 1)

    def _update_selected_preset(self) -> None:
        current_color = self.solid.color.lower()
        current_alpha = self.solid.alpha

        for button in self.preset_buttons:
            is_match = (button.color.lower() == current_color and
                       abs(button.alpha - current_alpha) < 0.01)
            button.set_selected(is_match)

        if self.color_picker_button:
            is_custom_match = (self.custom_color and
                             self.custom_color.lower() == current_color and
                             abs(self.custom_alpha - current_alpha) < 0.01)
            self.color_picker_button.set_selected(is_custom_match)

    def _on_color_picker_clicked(self, button: ColorPickerButton) -> None:
        button.open_color_picker()

    def _on_custom_color_picked(self, color: str, alpha: float) -> None:
        self.custom_color = color
        self.custom_alpha = alpha
        self.solid.color = color
        self.solid.alpha = alpha
        self._update_selected_preset()
        if self.callback:
            self.callback(self.solid)

    def _on_common_color_clicked(self, _button: Gtk.Button, color: str, alpha: float) -> None:
        self.solid.color = color
        self.solid.alpha = alpha
        self._update_selected_preset()
        if self.callback:
            self.callback(self.solid)
